# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from neurst.utils.activations import glu


class LightConvolutionLayer(tf.keras.layers.Layer):

    def __init__(self,
                 kernel_size,
                 num_heads,
                 conv_type="lightweight",
                 conv_dim=None,
                 use_glu=True,
                 weight_dropout_rate=0.,
                 name=None):
        super(LightConvolutionLayer, self).__init__(name=name)
        self._conv_type = conv_type
        self._conv_dim = conv_dim
        self._kernel_size = kernel_size
        self._num_heads = num_heads
        self._use_glu = use_glu
        self._weight_dropout_rate = weight_dropout_rate

    def get_config(self):
        return dict(
            conv_type=self._conv_type,
            kernel_size=self._kernel_size,
            num_heads=self._num_heads,
            conv_dim=self._conv_dim,
            use_glu=self._use_glu,
            weight_dropout_rate=self._weight_dropout_rate,
            name=self.name)

    def build(self, input_shape):
        input_dim = input_shape[-1]
        if self._conv_dim is None:
            self._conv_dim = input_dim
        assert self._conv_dim % self._num_heads == 0, (
            "The internal dimension({}) of conv must be divided "
            "evenly by num_heads({}).".format(self._conv_dim, self._num_heads))
        if self._use_glu:
            self._dense1 = tf.keras.layers.Dense(
                self._conv_dim * 2, activation=glu, use_bias=True, name="dense1")
        else:
            self._dense1 = tf.keras.layers.Dense(
                self._conv_dim, activation=None, use_bias=True, name="dense1")
        self._dense2 = tf.keras.layers.Dense(
            input_dim, activation=None, use_bias=True, name="dense2")
        if self._conv_type == "lightweight":
            self._conv_shared_weight = self.add_weight(
                "conv_shared_weight",
                shape=[self._num_heads, self._kernel_size],
                trainable=True)
        elif self._conv_type == "dynamic":
            self._conv_weight_linear = tf.keras.layers.Dense(
                self._num_heads * self._kernel_size, name="conv_weight_linear",
                use_bias=False, activation=None)
        else:
            raise NotImplementedError("Not supported conv_type: {}".format(self._conv_type))

    def dynamic_conv(self, x, left_padding=False, cache=None,
                     is_training=False, decode_loop_step=None):
        x_dim = x.get_shape().as_list()[-1]
        batch = tf.shape(x)[0]
        timestep = tf.shape(x)[1]
        # weight softmax [batch, time, heads, kernel]
        w = tf.reshape(self._conv_weight_linear(x),
                       [batch, timestep, self._num_heads, self._kernel_size])
        w = tf.nn.softmax(w)
        if left_padding:
            if cache is not None:
                if decode_loop_step is None:
                    x = tf.concat([cache["conv"], x], axis=1)
                    cache["conv"] = x
                    # [batch, timesteps(kernel_size), dim]
                    x = x[:, -self._kernel_size:, :]
                else:
                    _cache = cache["conv"]
                    size = _cache.get_shape().as_list()[1]
                    indices = tf.reshape(
                        tf.one_hot(decode_loop_step + self._kernel_size - 1, size, dtype=x.dtype),
                        [1, size, 1])
                    cache["conv"] = _cache + x * indices
                    # [batch, timesteps(kernel_size), dim]
                    x = cache["conv"][:, decode_loop_step:(self._kernel_size + decode_loop_step), :]
                # [batch, 1, dim, kernel_size]
                x_unfold = tf.expand_dims(tf.transpose(x, [0, 2, 1]), 1)
            else:
                x = tf.pad(x, [[0, 0], [self._kernel_size - 1, 0], [0, 0]], mode="CONSTANT",
                           constant_values=0)
                # x: [batch, time, dim] => [batch, 1, time, dim]
                x_unfold = tf.image.extract_patches(
                    images=tf.cast(tf.expand_dims(x, 1), dtype=tf.float32),
                    sizes=[1, 1, self._kernel_size, 1],
                    strides=[1, 1, 1, 1],
                    rates=[1, 1, 1, 1],
                    padding="VALID")
                # [batch, time, dim, kernel]
                x_unfold = tf.transpose(tf.reshape(tf.cast(x_unfold, dtype=x.dtype),
                                                   [batch, timestep, self._kernel_size, -1]), [0, 1, 3, 2])
        else:
            # x: [batch, time, dim] => [batch, 1, time, dim]
            x_unfold = tf.image.extract_patches(
                images=tf.cast(tf.expand_dims(x, 1), dtype=tf.float32),
                sizes=[1, 1, self._kernel_size, 1],
                strides=[1, 1, 1, 1],
                rates=[1, 1, 1, 1],
                padding="SAME")
            # [batch, time, dim, kernel]
            x_unfold = tf.transpose(tf.reshape(tf.cast(x_unfold, dtype=x.dtype),
                                               [batch, timestep, self._kernel_size, -1]), [0, 1, 3, 2])

        if is_training:
            w = tf.nn.dropout(w, rate=self._weight_dropout_rate)
        # [batch, time, heads, dim//heads, kernel]
        x_unfold = tf.reshape(x_unfold, [batch, timestep, self._num_heads, x_dim // self._num_heads,
                                         self._kernel_size])
        conv_out = tf.reshape(tf.einsum("abcde,abce->abcd", x_unfold, w), [batch, timestep, x_dim])
        return conv_out

    def lightweight_conv(self, x, left_padding=False, cache=None,
                         is_training=False, decode_loop_step=None):
        """ Use tf.raw_ops. """
        x_dim = x.get_shape().as_list()[-1]
        if left_padding:
            if cache is None:  # for decoder under training
                x = tf.pad(x, [[0, 0], [self._kernel_size - 1, 0], [0, 0]], mode="CONSTANT",
                           constant_values=0)
            else:  # for decoder under inference
                if decode_loop_step is None:
                    x = tf.concat([cache["conv"], x], axis=1)
                    cache["conv"] = x
                    # [batch, timesteps(kernel_size), dim]
                    x = cache["conv"][:, -self._kernel_size:, :]
                else:
                    _cache = cache["conv"]
                    size = _cache.get_shape().as_list()[1]
                    indices = tf.reshape(
                        tf.one_hot(decode_loop_step + self._kernel_size - 1, size, dtype=x.dtype),
                        [1, size, 1])
                    cache["conv"] = _cache + x * indices
                    # [batch, timesteps(kernel_size), dim]
                    x = cache["conv"][:, decode_loop_step:(self._kernel_size + decode_loop_step), :]
            padding_ = "VALID"
        else:
            padding_ = "SAME"

        # [batch, time, 1, dim]
        x = tf.expand_dims(x, axis=2)
        # weight softmax: [heads, kernel]
        w = tf.nn.softmax(self._conv_shared_weight)
        # [kernel, dim]
        filter = tf.reshape(tf.repeat(tf.transpose(w), repeats=x_dim // self._num_heads),
                            [self._kernel_size, x_dim])
        # [kernel, 1, dim, 1]
        filter = tf.expand_dims(tf.expand_dims(filter, 1), 3)
        if is_training:
            filter = tf.nn.dropout(filter, rate=self._weight_dropout_rate)
        # [batch, time, 1, dim]
        conv_out = tf.raw_ops.DepthwiseConv2dNative(
            input=x, filter=filter, strides=[1, 1, 1, 1], padding=padding_)
        conv_out = tf.squeeze(conv_out, axis=2)
        return conv_out

    def call(self, inputs, inputs_padding=None, left_padding=False,
             cache=None, is_training=False, decode_loop_step=None):
        """

        Args:
            inputs: A float tensor of shape [batch_size, length, num_units].
            inputs_padding: A float tensor with the same as `inputs`, where 1.0 indicates the padding position.
            left_padding: Whether to pad (kernel_size-1) tokens on the left of length-dim.
            cache: The internal cache for inference.
            is_training: A bool, whether in training mode or not.
            decode_loop_step: An integer, step number of the decoding loop. Used only
                for autoregressive inference with static-shape cache.

        Returns: A float tensor of shape [batch_size, length, num_units]
        """
        x = self._dense1(inputs)
        if inputs_padding is not None:
            x *= tf.expand_dims((1. - inputs_padding), 2)
        if self._conv_type == "lightweight":
            conv_output = self.lightweight_conv(x,
                                                left_padding=left_padding,
                                                cache=cache, is_training=is_training,
                                                decode_loop_step=decode_loop_step)
        elif self._conv_type == "dynamic":
            conv_output = self.dynamic_conv(x,
                                            left_padding=left_padding,
                                            cache=cache, is_training=is_training,
                                            decode_loop_step=decode_loop_step)
        else:
            raise NotImplementedError("Not supported conv_type: {}".format(self._conv_type))
        return self._dense2(conv_output)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Tuple

import tensorflow as tf
from absl import logging

import neurst.data.dataset_utils as dataset_utils
from neurst.data.data_pipelines import DataPipeline, build_data_pipeline
from neurst.data.data_pipelines.text_data_pipeline import TextDataPipeline
from neurst.data.datasets import Dataset
from neurst.data.datasets.parallel_text_dataset import AbstractParallelDataset
from neurst.layers.metric_layers.token_metric_layers import BatchCountMetricLayer, SequenceTokenMetricLayer
from neurst.metrics import build_metric
from neurst.models import build_model
from neurst.models.model_utils import deduce_text_length
from neurst.tasks import register_task
from neurst.tasks.task import Task
from neurst.utils import compat
from neurst.utils.configurable import deep_merge_dict
from neurst.utils.flags_core import Flag, ModuleFlag


@register_task("seq_to_seq")
class Seq2Seq(Task):
    """ Defines the sequence to sequence task. """

    def __init__(self, args):
        """ Initializes the task.

        Args:
            args: A dict of model configurations.
        """
        src_data_pipeline_cls = args.get("src_data_pipeline.class", TextDataPipeline)
        src_data_pipeline_params = args.get("src_data_pipeline.params", None) or {}
        self._src_data_pipeline = build_data_pipeline(
            src_data_pipeline_cls, **src_data_pipeline_params)
        trg_data_pipeline_cls = args.get("trg_data_pipeline.class", TextDataPipeline)
        trg_data_pipeline_params = args.get("trg_data_pipeline.params", None) or {}
        self._trg_data_pipeline = build_data_pipeline(
            trg_data_pipeline_cls, **trg_data_pipeline_params)
        self._target_begin_of_sentence = args["target_begin_of_sentence"]
        super(Seq2Seq, self).__init__(args)

    def get_config(self):
        return {
            "src_data_pipeline.class": self._src_data_pipeline.__class__.__name__,
            "src_data_pipeline.params": self._src_data_pipeline.get_config(),
            "trg_data_pipeline.class": self._trg_data_pipeline.__class__.__name__,
            "trg_data_pipeline.params": self._trg_data_pipeline.get_config(),
            "target_begin_of_sentence": self._target_begin_of_sentence
        }

    @staticmethod
    def class_or_method_args():
        this_args = super(Seq2Seq, Seq2Seq).class_or_method_args()
        this_args.extend([
            # for creating data pipelines
            ModuleFlag("src_data_pipeline", DataPipeline.REGISTRY_NAME,
                       help="The source side data pipeline."),
            ModuleFlag("trg_data_pipeline", DataPipeline.REGISTRY_NAME,
                       help="The target side data pipeline."),
            # for preprocessing data
            Flag("max_src_len", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The maximum source length of training data."),
            Flag("max_trg_len", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The maximum target length of training data."),
            Flag("truncate_src", dtype=Flag.TYPE.BOOLEAN, default=None,
                 help="Whether to truncate source to max_src_len."),
            Flag("truncate_trg", dtype=Flag.TYPE.BOOLEAN, default=None,
                 help="Whether to truncate target to max_trg_len."),
            # for batching dataset
            Flag("batch_by_tokens", dtype=Flag.TYPE.BOOLEAN, default=None,
                 help="Whether to batch the data by word tokens."),
            Flag("target_begin_of_sentence", dtype=Flag.TYPE.STRING, default="bos",
                 choices=["bos", "eos"],
                 help="The begin of sentence symbol for target side. The choice 'eos' "
                      "is for compatibility with fairseq transformer.")
        ])
        return this_args

    def inputs_signature(self, mode) -> Tuple[dict, dict]:
        """ Returns the input dtypes and signatures. """
        dtypes = {"feature": tf.int64}
        signatures = {"feature": tf.TensorShape([None, None])}
        if mode == compat.ModeKeys.INFER:
            return dtypes, signatures
        dtypes["label"] = tf.int64
        signatures["label"] = tf.TensorShape([None, None])
        return dtypes, signatures

    def build_model(self, args, name=None):
        """ Builds and return a keras model. """
        model = build_model(args, self._src_data_pipeline.meta,
                            self._trg_data_pipeline.meta, name=name)
        return model

    def example_to_input(self, batch_of_data: dict, mode) -> dict:
        """ Transform the data examples to model acceptable inputs.

        Args:
            batch_of_data: A data tensor with shape [batch, ...]
            mode: The running mode.

        Returns: The input data for model.
        """
        input_dict = {"src": batch_of_data["feature"],
                      "src_length": deduce_text_length(batch_of_data["feature"],
                                                       self._src_data_pipeline.meta["pad_id"],
                                                       self._src_data_pipeline.meta["padding_mode"])}
        bosid = (self._trg_data_pipeline.meta["eos_id"] if self._target_begin_of_sentence == "eos"
                 else self._trg_data_pipeline.meta["bos_id"])
        target_bos = tf.tile([tf.convert_to_tensor(bosid, dtype=tf.int64)],
                             [tf.shape(input_dict["src"])[0]])
        if mode == compat.ModeKeys.INFER:
            input_dict["trg_input"] = target_bos
        else:
            input_dict["trg"] = batch_of_data["label"]
            input_dict["trg_length"] = deduce_text_length(batch_of_data["label"],
                                                          self._trg_data_pipeline.meta["pad_id"],
                                                          self._trg_data_pipeline.meta["padding_mode"])
            input_dict["trg_input"] = tf.concat([tf.expand_dims(target_bos, axis=1),
                                                 batch_of_data["label"][:, :-1]], axis=1)
        return input_dict

    def get_data_postprocess_fn(self, mode):
        if mode == compat.ModeKeys.INFER:
            return self._trg_data_pipeline.recover
        raise ValueError("No postprocess for TRAIN/EVAL.")

    def get_data_preprocess_fn(self, mode, data_status=compat.DataStatus.RAW, args=None) -> callable:
        """ Preprocess data sample according to this task.

        Args:
            args: A dict containing dataset arguments.
            mode: A ModeKeys indicating the running mode.
            data_status: The status of the data sample.

        Returns: A callable function to collate (process) a data sample.
        """
        if args is None:
            args = self._args
        else:
            args = deep_merge_dict(self._args, args, local_overwrite=False)
        truncate_src = args.get("truncate_src", None)
        truncate_trg = args.get("truncate_trg", None)
        max_src_len = args.get("max_src_len", None)
        max_trg_len = args.get("max_trg_len", None)

        def _process_and_truncate(text, dp, trunc, max_len):
            if data_status != compat.DataStatus.PROJECTED:
                text = dp.process(text, is_processed=(data_status == compat.DataStatus.PROCESSED))
            if mode == compat.ModeKeys.TRAIN and trunc and max_len:
                if isinstance(text, tf.Tensor):
                    text = tf.cond(
                        tf.less_equal(tf.size(text), max_len), lambda: text,
                        lambda: tf.concat([text[:(max_len - 1)], text[-1:]], axis=0))
                elif len(text) > max_len:
                    text = text[:(max_len - 1)] + text[-1:]
            return text

        if mode == compat.ModeKeys.INFER:
            return lambda data: {
                "feature": _process_and_truncate(data["feature"],
                                                 self._src_data_pipeline,
                                                 truncate_src,
                                                 max_src_len)}
        return lambda data: {
            "feature": _process_and_truncate(data["feature"],
                                             self._src_data_pipeline,
                                             truncate_src,
                                             max_src_len),
            "label": _process_and_truncate(data["label"],
                                           self._trg_data_pipeline,
                                           truncate_trg,
                                           max_trg_len)}

    def create_and_batch_tfds(self, ds: Dataset, mode,
                              args=None, num_replicas_in_sync=1) -> tf.data.Dataset:
        """ Creates a dataset according to the `mode`.

        Args:
            args: A dict containing dataset arguments.
            ds: A neurst.data.datasets.Dataset object.
            mode: A ModeKeys indicating the running mode.
            num_replicas_in_sync: The number of GPUs or other workers. We will generate global
                batches, and each global batch is equally divisible by number of replicas.

        Returns:
            A tf.data.Dataset.
        """
        if args is None:
            args = self._args
        else:
            args = deep_merge_dict(self._args, args, local_overwrite=False)
        src_eos = tf.constant(self._src_data_pipeline.meta["eos_id"], dtype=tf.int64)
        trg_eos = tf.constant(self._trg_data_pipeline.meta["eos_id"], dtype=tf.int64)

        assert isinstance(ds, AbstractParallelDataset), (
            "The dataset for SeqToSeq task must inherit AbstractParallelDataset.")

        dataset = ds.build(map_func=self.get_data_preprocess_fn(mode, ds.status, args),
                           map_output_dtypes=self.inputs_signature(mode)[0],
                           auto_shard=(mode == compat.ModeKeys.TRAIN),
                           shuffle=(mode == compat.ModeKeys.TRAIN))

        if mode == compat.ModeKeys.INFER:
            logging.info("Creating test dataset.")
            return dataset.cache().padded_batch(
                dataset_utils.adjust_batch_size(args["batch_size"],
                                                num_replicas_in_sync=num_replicas_in_sync),
                padded_shapes={"feature": [None]},
                padding_values={"feature": src_eos},
                drop_remainder=False)
        elif mode == compat.ModeKeys.EVAL:
            logging.info("Creating evaluation dataset.")
            return dataset.cache().padded_batch(
                dataset_utils.adjust_batch_size(args["batch_size"],
                                                num_replicas_in_sync=num_replicas_in_sync),
                padded_shapes={"feature": [None], "label": [None]},
                padding_values={"feature": src_eos, "label": trg_eos},
                drop_remainder=False)
        else:
            logging.info("Creating training dataset.")
            dataset = dataset_utils.clean_dataset_by_length(
                dataset, {"feature": args["max_src_len"], "label": args["max_trg_len"]})
            if args["cache_dataset"]:
                dataset = dataset.cache()
            if args["shuffle_buffer"]:
                dataset = dataset.shuffle(buffer_size=args["shuffle_buffer"])
            padding_values = {"feature": src_eos, "label": trg_eos}
            if args["max_src_len"] is None:
                raise RuntimeError("Must provide `max_src_len` for training.")
            if args["max_trg_len"] is None:
                raise RuntimeError("Must provide `max_trg_len` for training.")
            src_bucket_boundaries, trg_bucket_boundaries = dataset_utils.associated_bucket_boundaries(
                dataset_utils.create_batch_bucket_boundaries(args["max_src_len"]),
                dataset_utils.create_batch_bucket_boundaries(args["max_trg_len"]))

            bucket_boundaries = {
                "feature": src_bucket_boundaries,
                "label": trg_bucket_boundaries
            }
            bucket_batch_sizes = dataset_utils.adjust_batch_size(
                args["batch_size"],
                args["batch_size_per_gpu"],
                bucket_boundaries=bucket_boundaries if args["batch_by_tokens"] else None,
                boundaries_reduce_to_length_fn=lambda x: max(tf.nest.flatten(x)),
                num_replicas_in_sync=num_replicas_in_sync)
            return dataset_utils.batch_examples_by_token(
                dataset,
                bucket_boundaries=bucket_boundaries,
                bucket_batch_sizes=bucket_batch_sizes,
                padding_values=padding_values,
                example_length_func=lambda x: {k: tf.size(v) for k, v in x.items()}
            )

    def build_metric_layer(self):
        return [SequenceTokenMetricLayer("src"), SequenceTokenMetricLayer("trg"),
                BatchCountMetricLayer("src")]

    def get_eval_metric(self, args, name="metric"):
        """ Returns a neurst.metrics.metric.Metric object for evaluation."""
        return build_metric(args[name + ".class"], language=self._trg_data_pipeline.meta["language"],
                            **args[name + ".params"])



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from absl import app, logging

import neurst.utils.flags_core as flags_core
from neurst.metrics import Metric, build_metric

FLAG_LIST = [
    flags_core.Flag("hypo_file", dtype=flags_core.Flag.TYPE.STRING, default=None,
                    help="The path to hypothesis file."),
    flags_core.Flag("ref_file", dtype=flags_core.Flag.TYPE.STRING, default=None,
                    help="The path to reference file."),
    flags_core.ModuleFlag(Metric.REGISTRY_NAME, help="The metric for evaluation."),
]


def evaluate(metric, hypo_file, ref_file):
    assert metric is not None
    assert hypo_file
    assert ref_file
    with tf.io.gfile.GFile(hypo_file) as fp:
        hypo = [line.strip() for line in fp]
    with tf.io.gfile.GFile(ref_file) as fp:
        ref = [line.strip() for line in fp]

    metric_result = metric(hypo, ref)
    for k, v in metric_result.items():
        logging.info("Evaluation result: %s=%.2f", k, v)


def _main(_):
    arg_parser = flags_core.define_flags(FLAG_LIST, with_config_file=False)
    args, remaining_argv = flags_core.intelligent_parse_flags(FLAG_LIST, arg_parser)
    flags_core.verbose_flags(FLAG_LIST, args, remaining_argv)
    metric = build_metric(args)
    evaluate(metric, args["hypo_file"], args["ref_file"])


if __name__ == "__main__":
    logging.set_verbosity(logging.INFO)
    app.run(_main, argv=["pseudo.py"])



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import abstractmethod

import tensorflow as tf

from neurst.layers.metric_layers import METRIC_REDUCTION, register_metric


class MetricLayer(tf.keras.layers.Layer):
    """ The base class of metric layer for verbose and """

    def __init__(self):
        super(MetricLayer, self).__init__()
        self._layer_metrics = {}

    def build_metric_reduction(self, name, reduction):
        register_metric(name, reduction)
        if reduction == METRIC_REDUCTION.SUM:
            self._layer_metrics[name] = tf.keras.metrics.Sum(name)
        elif reduction == METRIC_REDUCTION.MEAN:
            self._layer_metrics[name] = tf.keras.metrics.Mean(name)
        else:
            raise NotImplementedError(f"Unknown reduction name: {reduction}.")

    @abstractmethod
    def calculate(self, input, output):
        """ Calculates metric values according to model input and output. """
        raise NotImplementedError

    def call(self, inputs):
        """ Registers metrics by calling `self.add_metric()`

        Args:
            inputs: A list of [model inputs, model outputs]

        Returns:
            The model outputs.
        """
        model_inp, model_out = inputs
        ms = self.calculate(model_inp, model_out)
        if not isinstance(ms, dict):
            assert len(self._layer_metrics) == 1, "The number of metrics mismatch."
            for k, v in self._layer_metrics.items():
                ms = {k: ms}
        for name, aggr in self._layer_metrics.items():
            m = aggr(ms[name])
            self.add_metric(m)
        return model_out



from collections import namedtuple

METRIC_REDUCTION = namedtuple(
    "metric_reduction", "SUM MEAN")(0, 1)

REGISTERED_METRICS = dict()


def register_metric(name, redution):
    if name in REGISTERED_METRICS:
        raise ValueError(f"Metric {name} already registered.")
    REGISTERED_METRICS[name] = redution


def get_metric_reduction(name, default=METRIC_REDUCTION.MEAN):
    return REGISTERED_METRICS.get(name, default)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

import tensorflow as tf
from absl import logging

from neurst.data.text.vocab import PaddingMode
from neurst.utils import compat


def deduce_text_length(data_tensor, pad_id, padding_mode):
    """ Gets the length by text data tensor.

    Args:
        data_tensor: A Tensor with shape [batch_size, max_len]
        pad_id: The id of the PAD symbol.
        padding_mode: A PaddingMode value.

    Returns:
        The length Tensor with shape [batch_size, ]
    """
    if padding_mode == PaddingMode.DEFAULT:
        return tf.reduce_sum(tf.cast(
            tf.not_equal(data_tensor, pad_id), tf.int32), axis=1)
    elif padding_mode == PaddingMode.EOS_AS_PADDING:
        return tf.reduce_sum(tf.cast(
            tf.not_equal(data_tensor, pad_id), tf.int32), axis=1) + 1
    else:
        raise NotImplementedError


def input_length_to_nonpadding(lengths, max_len, dtype=None):
    """ Creates a bias tensor according to the non-padding tensor for cross entropy.

    Args:
        length: A Tensor with shape [batch_size, ], indicating the true length.
        max_len: A scalar tensor indicating the maximum length.

    Returns:
        A float tensor with shape [batch_size, max_len],
        indicating the padding positions, where 0.0 for padding and
        1.0 for non-padding.
    """
    return tf.sequence_mask(
        lengths=tf.cast(lengths, tf.int32),
        maxlen=tf.cast(max_len, tf.int32),
        dtype=(dtype or tf.dtypes.as_dtype(compat.CUSTOM_GLOBAL_FLOATX)))  # 1.0 for padding


def input_length_to_padding(lengths, max_len, dtype=None):
    """ Creates a bias tensor according to the padding tensor for attention.

    Args:
        length: A Tensor with shape [batch_size, ], indicating the true length.
        max_len: A scalar tensor indicating the maximum length.

    Returns:
        A float tensor with shape [batch_size, max_len],
        indicating the padding positions, where 1.0 for padding and
        0.0 for non-padding.
    """
    with tf.name_scope("padding"):
        return 1. - input_length_to_nonpadding(lengths, max_len, dtype)


def _summary_model_variables(varname_shape_list, freeze_variables=None,
                             output_stream=logging.info):
    output_stream("variable name  | # parameters")
    var_sizes = dict()
    var_hiro_names = dict()
    var_trainable = dict()
    non_trainable_sizes = 0
    for var_name, var_shape_list, trainable in varname_shape_list:
        if (trainable and freeze_variables is not None
            and re.search(freeze_variables, var_name) is not None):
            trainable = False
        var_names = var_name.split(":")[0].split("/")
        var_total_size = 1
        for _s in var_shape_list:
            var_total_size *= _s
        if var_names[0] not in var_hiro_names:
            var_hiro_names[var_names[0]] = dict()
        _tmp_var_names = var_hiro_names[var_names[0]]
        for i in range(1, len(var_names)):
            _tmp_name = "/".join(var_names[:i])
            if _tmp_name not in var_sizes:
                var_sizes[_tmp_name] = 0
            var_sizes[_tmp_name] += var_total_size
            if i == len(var_names) - 1:
                _tmp_var_names[var_names[i]] = var_shape_list
            else:
                if var_names[i] not in _tmp_var_names:
                    _tmp_var_names[var_names[i]] = dict()
                _tmp_var_names = _tmp_var_names[var_names[i]]
        var_sizes["/".join(var_names)] = var_total_size
        var_trainable["/".join(var_names)] = trainable
        if not trainable:
            non_trainable_sizes += var_total_size

    def _get_size(v):
        killo = v / 1000.
        milli = v / 1000000.
        if milli >= 1:
            return "%.2fm" % milli
        elif killo >= 1:
            return "%.2fk" % killo
        return v

    def _recursive_print(accumulate_name_list, recuv_varname_dict, indent=0):
        for k, v in recuv_varname_dict.items():
            accumulate_name_list.append(k)
            prefix = " "
            if indent > 0:
                prefix = "".join([" "] * (indent + 1))
            this_name = "/".join(accumulate_name_list)
            if isinstance(v, dict):
                extra = (f", {_get_size(var_sizes[this_name]-non_trainable_sizes)} trainable params"
                         if non_trainable_sizes > 0 and indent == 0 else "")
                output_stream(prefix + this_name + " (--/{} params{})".format(
                    _get_size(var_sizes[this_name]), extra))
                _recursive_print(accumulate_name_list, v, indent + 2)
            else:
                extra = "" if var_trainable[this_name] else ", non-trainable"
                output_stream(prefix + this_name + " ({}, {} params{})".format(
                    "x".join([str(_v) for _v in v]), _get_size(var_sizes[this_name]), extra))
            accumulate_name_list.pop(-1)

    _recursive_print([], var_hiro_names)


def summary_model_variables(keras_model, freeze_variables=None):
    """ Prints model variables and shapes.

    Args:
        keras_model: A keras model.
    """
    varname_shape_list = [
        (var.name, var.shape.as_list(), var.trainable) for var in keras_model.weights]
    _summary_model_variables(varname_shape_list, freeze_variables)



import collections
import re
import sys
import unicodedata

import six
import tensorflow as tf
from absl import logging

from neurst.data.text import register_tokenizer
from neurst.data.text.tokenizer import Tokenizer

RESERVED_TOKENS = []

# Regular expression for unescaping token strings.
# '\u' is converted to '_'
# '\\' is converted to '\'
# '\213;' is converted to unichr(213)
_UNESCAPE_REGEX = re.compile(r"\\u|\\\\|\\([0-9]+);")
_ESCAPE_CHARS = set(u"\\_u;0123456789")

# This set contains all letter (L) & number (N) of unicode chars.
_ALPHANUMERIC_CHAR_SET = set(
    six.unichr(i) for i in range(sys.maxunicode)
    if (unicodedata.category(six.unichr(i)).startswith("L")
        or unicodedata.category(six.unichr(i)).startswith("N")))

# min_count is the minimum number of times a subtoken must appear in the data
# before before it is added to the vocabulary. The value is found using binary
# search to obtain the target vocabulary size.
_MIN_MIN_COUNT = 1  # min value to use when binary searching for min_count
_MAX_MIN_COUNT = 1000  # max value to use when binary searching for min_count


def _generate_alphabet_dict(iterable, reserved_tokens=None):
    """Create set of characters that appear in any element in the iterable."""
    if reserved_tokens is None:
        reserved_tokens = RESERVED_TOKENS
    alphabet = {c for token in iterable for c in token}
    alphabet |= {c for token in reserved_tokens for c in token}
    alphabet |= _ESCAPE_CHARS  # Add escape characters to alphabet set.
    return alphabet


def _save_vocab_file(vocab_file, subtoken_list):
    """Save subtokens to file."""
    with tf.io.gfile.GFile(vocab_file, mode="w") as f:
        for subtoken in subtoken_list:
            f.write("'%s'\n" % subtoken)


def _list_to_index_dict(lst):
    """Create dictionary mapping list items to their indices in the list."""
    return {item: n for n, item in enumerate(lst)}


def _split_token_to_subtokens(token, subtoken_dict, max_subtoken_length):
    """Splits a token into subtokens defined in the subtoken dict."""
    ret = []
    start = 0
    token_len = len(token)
    while start < token_len:
        # Find the longest subtoken, so iterate backwards.
        for end in range(min(token_len, start + max_subtoken_length), start, -1):
            subtoken = token[start:end]
            if subtoken in subtoken_dict:
                ret.append(subtoken)
                start = end
                break
        else:  # Did not break
            # If there is no possible encoding of the escaped token then one of the
            # characters in the token is not in the alphabet. This should be
            # impossible and would be indicative of a bug.
            raise ValueError("Was unable to split token \"%s\" into subtokens." %
                             token)
    return ret


def _count_and_gen_subtokens(token_counts, alphabet, subtoken_dict,
                             max_subtoken_length):
    """Count number of times subtokens appear, and generate new subtokens.

    Args:
      token_counts: dict mapping tokens to the number of times they appear in the
        original files.
      alphabet: list of allowed characters. Used to escape the tokens, which
        guarantees that all tokens can be split into subtokens.
      subtoken_dict: dict mapping subtokens to ids.
      max_subtoken_length: maximum length of subtoken in subtoken_dict.

    Returns:
      A defaultdict mapping subtokens to the number of times they appear in the
      tokens. The dict may contain new subtokens.
    """
    subtoken_counts = collections.defaultdict(int)
    for token, count in six.iteritems(token_counts):
        token = _escape_token(token, alphabet)
        subtokens = _split_token_to_subtokens(token, subtoken_dict,
                                              max_subtoken_length)

        # Generate new subtokens by taking substrings from token.
        start = 0
        for subtoken in subtokens:
            for end in range(start + 1, len(token) + 1):
                new_subtoken = token[start:end]
                subtoken_counts[new_subtoken] += count
            start += len(subtoken)

    return subtoken_counts


def _filter_and_bucket_subtokens(subtoken_counts, min_count):
    """Return a bucketed list of subtokens that are filtered by count.

    Args:
      subtoken_counts: defaultdict mapping subtokens to their counts
      min_count: int count used to filter subtokens

    Returns:
      List of subtoken sets, where subtokens in set i have the same length=i.
    """
    # Create list of buckets, where subtokens in bucket i have length i.
    subtoken_buckets = []
    for subtoken, count in six.iteritems(subtoken_counts):
        if count < min_count:  # Filter out subtokens that don't appear enough
            continue
        while len(subtoken_buckets) <= len(subtoken):
            subtoken_buckets.append(set())
        subtoken_buckets[len(subtoken)].add(subtoken)
    return subtoken_buckets


def _gen_new_subtoken_list(subtoken_counts,
                           min_count,
                           alphabet,
                           reserved_tokens=None):
    """Generate candidate subtokens ordered by count, and new max subtoken length.

    Add subtokens to the candiate list in order of length (longest subtokens
    first). When a subtoken is added, the counts of each of its prefixes are
    decreased. Prefixes that don't appear much outside the subtoken are not added
    to the candidate list.

    For example:
      subtoken being added to candidate list: 'translate'
      subtoken_counts: {'translate':10, 't':40, 'tr':16, 'tra':12, ...}
      min_count: 5

    When 'translate' is added, subtoken_counts is updated to:
      {'translate':0, 't':30, 'tr':6, 'tra': 2, ...}

    The subtoken 'tra' will not be added to the candidate list, because it appears
    twice (less than min_count) outside of 'translate'.

    Args:
      subtoken_counts: defaultdict mapping str subtokens to int counts
      min_count: int minumum count requirement for subtokens
      alphabet: set of characters. Each character is added to the subtoken list to
        guarantee that all tokens can be encoded.
      reserved_tokens: list of tokens that will be added to the beginning of the
        returned subtoken list.

    Returns:
      List of candidate subtokens in decreasing count order, and maximum subtoken
      length
    """
    if reserved_tokens is None:
        reserved_tokens = RESERVED_TOKENS

    # Create a list of (count, subtoken) for each candidate subtoken.
    subtoken_candidates = []

    # Use bucketted list to iterate through subtokens in order of length.
    # subtoken_buckets[i] = set(subtokens), where each subtoken has length i.
    subtoken_buckets = _filter_and_bucket_subtokens(subtoken_counts, min_count)
    max_subtoken_length = len(subtoken_buckets) - 1

    # Go through the list in reverse order to consider longer subtokens first.
    for subtoken_len in range(max_subtoken_length, 0, -1):
        for subtoken in subtoken_buckets[subtoken_len]:
            count = subtoken_counts[subtoken]

            # Possible if this subtoken is a prefix of another token.
            if count < min_count:
                continue

            # Ignore alphabet/reserved tokens, which will be added manually later.
            if subtoken not in alphabet and subtoken not in reserved_tokens:
                subtoken_candidates.append((count, subtoken))

            # Decrement count of the subtoken's prefixes (if a longer subtoken is
            # added, its prefixes lose priority to be added).
            for end in range(1, subtoken_len):
                subtoken_counts[subtoken[:end]] -= count

    # Add alphabet subtokens (guarantees that all strings are encodable).
    subtoken_candidates.extend((subtoken_counts.get(a, 0), a) for a in alphabet)

    # Order subtoken candidates by decreasing count.
    subtoken_list = [t for _, t in sorted(subtoken_candidates, reverse=True)]

    # Add reserved tokens to beginning of the list.
    subtoken_list = reserved_tokens + subtoken_list
    return subtoken_list, max_subtoken_length


def _generate_subtokens(token_counts,
                        alphabet,
                        min_count,
                        num_iterations=4,
                        reserved_tokens=None):
    """Create a list of subtokens in decreasing order of frequency.

    Args:
      token_counts: dict mapping str tokens -> int count
      alphabet: set of characters
      min_count: int minimum number of times a subtoken must appear before it is
        added to the vocabulary.
      num_iterations: int number of iterations to generate new tokens.
      reserved_tokens: list of tokens that will be added to the beginning to the
        returned subtoken list.

    Returns:
      Sorted list of subtokens (most frequent first)
    """
    if reserved_tokens is None:
        reserved_tokens = RESERVED_TOKENS

    # Use alphabet set to create initial list of subtokens
    subtoken_list = reserved_tokens + list(alphabet)
    max_subtoken_length = 1

    # On each iteration, segment all words using the subtokens defined in
    # subtoken_dict, count how often the resulting subtokens appear, and update
    # the dictionary with subtokens w/ high enough counts.
    for i in range(num_iterations):
        logging.info("\tGenerating subtokens: iteration %d", i)
        # Generate new subtoken->id dictionary using the new subtoken list.
        subtoken_dict = _list_to_index_dict(subtoken_list)

        # Create dict mapping subtoken->count, with additional subtokens created
        # from substrings taken from the tokens.
        subtoken_counts = _count_and_gen_subtokens(token_counts, alphabet,
                                                   subtoken_dict,
                                                   max_subtoken_length)

        # Generate new list of subtokens sorted by subtoken count.
        subtoken_list, max_subtoken_length = _gen_new_subtoken_list(
            subtoken_counts, min_count, alphabet, reserved_tokens)

        logging.info("\tVocab size: %d", len(subtoken_list))
    return subtoken_list


def _generate_subtokens_with_target_vocab_size(token_counts,
                                               alphabet,
                                               target_size,
                                               threshold,
                                               min_count=None,
                                               reserved_tokens=None):
    """Generate subtoken vocabulary close to the target size."""
    if reserved_tokens is None:
        reserved_tokens = RESERVED_TOKENS

    if min_count is not None:
        logging.info("Using min_count=%d to generate vocab with target size %d",
                     min_count, target_size)
        return _generate_subtokens(
            token_counts, alphabet, min_count, reserved_tokens=reserved_tokens)

    def bisect(min_val, max_val):
        """Recursive function to binary search for subtoken vocabulary."""
        cur_count = (min_val + max_val) // 2
        logging.info("Binary search: trying min_count=%d (%d %d)", cur_count,
                     min_val, max_val)
        subtoken_list = _generate_subtokens(
            token_counts, alphabet, cur_count, reserved_tokens=reserved_tokens)

        val = len(subtoken_list)
        logging.info("Binary search: min_count=%d resulted in %d tokens", cur_count,
                     val)

        within_threshold = abs(val - target_size) < threshold
        if within_threshold or min_val >= max_val or cur_count < 2:
            return subtoken_list
        if val > target_size:
            other_subtoken_list = bisect(cur_count + 1, max_val)
        else:
            other_subtoken_list = bisect(min_val, cur_count - 1)

        # Return vocabulary dictionary with the closest number of tokens.
        other_val = len(other_subtoken_list)
        if abs(other_val - target_size) < abs(val - target_size):
            return other_subtoken_list
        return subtoken_list

    logging.info("Finding best min_count to get target size of %d", target_size)
    return bisect(_MIN_MIN_COUNT, _MAX_MIN_COUNT)


def _split_string_to_tokens(text, space_wrapper=None):
    """ The pre-tokenization for alphanums.

    Args:
        text: The input text.

    Returns: A list of tokens.
    """
    if not text:
        return []
    ret = []
    token_start = 0
    is_alnum = [c in _ALPHANUMERIC_CHAR_SET for c in text]
    for pos in range(1, len(text)):
        if is_alnum[pos] != is_alnum[pos - 1]:
            token = text[token_start:pos]
            if token != " " or token_start == 0:
                ret.append(token)
            token_start = pos
    final_token = text[token_start:]
    ret.append(final_token)
    if space_wrapper:
        return [i.replace(" ", space_wrapper) for i in ret]
    return ret


def _count_tokens(files, file_byte_limit=1e6):
    """Return token counts of words in the files.

    Samples file_byte_limit bytes from each file, and counts the words that appear
    in the samples. The samples are semi-evenly distributed across the file.

    Args:
      files: List of filepaths
      file_byte_limit: Max number of bytes that will be read from each file.

    Returns:
      Dictionary mapping tokens to the number of times they appear in the sampled
      lines from the files.
    """

    token_counts = collections.defaultdict(int)

    for filepath in files:
        with tf.io.gfile.GFile(filepath, mode="r") as reader:
            file_byte_budget = file_byte_limit
            counter = 0
            lines_to_skip = int(reader.size() / (file_byte_budget * 2))
            for line in reader:
                if counter < lines_to_skip:
                    counter += 1
                else:
                    if file_byte_budget < 0:
                        break
                    line = line.strip()
                    file_byte_budget -= len(line)
                    counter = 0

                    # Add words to token counts
                    for token in _split_string_to_tokens(line):
                        token_counts[token] += 1
    return token_counts


def _unescape_token(escaped_token):
    """Inverse of _escape_token().

    Args:
        escaped_token: a unicode string

    Returns:
        token: a unicode string
    """

    def match(m):
        if m.group(1) is None:
            return u"_" if m.group(0) == u"\\u" else u"\\"

        try:
            return six.unichr(int(m.group(1)))
        except (ValueError, OverflowError):
            return u"\u3013"  # Unicode for undefined character.

    trimmed = escaped_token[:-1] if escaped_token.endswith("_") \
        else escaped_token
    return _UNESCAPE_REGEX.sub(match, trimmed)


def _escape_token(token, alphabet):
    """Escape away underscores and OOV characters and append '_'.

    This allows the token to be expressed as the concatenation of a list
    of subtokens from the vocabulary. The underscore acts as a sentinel
    which allows us to invertibly concatenate multiple such lists.

    Args:
        token: A unicode string to be escaped.
        alphabet: A set of all characters in the vocabulary's alphabet.

    Returns:
        escaped_token: An escaped unicode string.

    Raises:
        ValueError: If the provided token is not unicode.
    """
    if not isinstance(token, six.text_type):
        raise ValueError("Expected string type for token, got %s" % type(token))

    token = token.replace(u"\\", u"\\\\").replace(u"_", u"\\u")
    ret = [c if c in alphabet and c != u"\n" else r"\%d;" % ord(c) for c in token]  # noqa
    return u"".join(ret) + "_"


@register_tokenizer("sub_tokenizer")
class Subtokenizer(Tokenizer):

    def __init__(self, language, glossaries=None, **kwargs):
        _ = kwargs
        super(Subtokenizer, self).__init__(language=language, glossaries=None)
        self._space_wrapper = "▁"

        self._reserved_tokens = []
        if isinstance(glossaries, list):
            for word in glossaries:
                self._reserved_tokens.append(word)
                self._reserved_tokens.append(word + "_")

        # Initialize the cache to empty.
        self._cache_size = 2 ** 20
        self._cache = [(None, None)] * self._cache_size
        self._built = False

    def init_subtokenizer(self, codes):
        subtoken_list = []
        with tf.io.gfile.GFile(codes, mode="r") as f:
            for line in f:
                subtoken = line.strip()
                if ((subtoken.startswith("'") and subtoken.endswith("'"))
                    or (subtoken.startswith('"') and subtoken.endswith('"'))):
                    subtoken = subtoken[1:-1]  # Remove surrounding single-quotes
                if subtoken in self._reserved_tokens:
                    continue
                if " " in subtoken:
                    self._space_wrapper = None
                subtoken_list.append(subtoken)
        self._all_subtoken_strings = subtoken_list + self._reserved_tokens
        # we remember the maximum length of any subtoken to avoid having to
        # check arbitrarily long strings.
        self._max_subtoken_len = max([len(s) for s in self._all_subtoken_strings])
        # Include all characters from all tokens in the alphabet to guarantee
        # any token can be encoded. Include all escaping characters.
        self._alphabet = {c for token in self._all_subtoken_strings for c in token}
        self._alphabet |= _ESCAPE_CHARS
        self._built = True

    def split_string_to_tokens(self, text):
        """ The pre-tokenization for alphanums.

        Args:
            text: The input text.

        Returns: A list of tokens.
        """
        return _split_string_to_tokens(text, self._space_wrapper)

    def join_tokens_to_string(self, tokens):
        """ The post-detokenization for alphanums.

        Args:
            tokens: The input text or a list of token strings.

        Returns: A string.
        """
        s = self._convert_to_list(tokens)
        if self._space_wrapper:
            s = [i.replace(self._space_wrapper, " ") for i in s]
        token_is_alnum = [t[0] in _ALPHANUMERIC_CHAR_SET for t in s]
        ret = []
        for i, token in enumerate(s):
            if i > 0 and token_is_alnum[i - 1] and token_is_alnum[i]:
                ret.append(" ")
            ret.append(token)
        return self._convert_to_str(ret, delimiter="")

    def _escaped_token_to_subtoken_strings(self, escaped_token):
        """ Converts an escaped token string to a list of subtoken strings.
          Note that this algorithm is greedy; it won't produce the "best"
        Args:
            escaped_token: An escaped token string.

        Returns: A list of subtoken strings.
        """
        ret = []
        start = 0
        token_len = len(escaped_token)
        while start < token_len:
            for end in range(min(token_len, start + self._max_subtoken_len), start, -1):
                subtoken = escaped_token[start:end]
                if subtoken in self._all_subtoken_strings:
                    ret.append(subtoken)
                    start = end
                    break
            else:  # Did not break
                # If there is no possible encoding of the escaped token then one
                # characters in the token is not in the alphabet. This should be
                # impossible and would be indicative of a bug.
                assert False, "Token substring not found in subtoken vocabulary"
        return ret

    def _token_to_subtoken_strings(self, token):
        """Converts a token to a list of subtoken stringds.

        Args:
            token: a string.
        Returns:
            a list of subword tokens.
        """
        cache_location = hash(token) % self._cache_size
        cache_key, cache_value = self._cache[cache_location]
        if cache_key == token:
            return cache_value
        escaped_token = _escape_token(token, self._alphabet)
        ret = self._escaped_token_to_subtoken_strings(escaped_token)
        self._cache[cache_location] = (token, ret)
        return ret

    def tokenize(self, text, return_str=False):
        if not self._built:
            raise ValueError("call `init_subtokenizer` at first to initialize the Subtokenizer.")
        words = self.split_string_to_tokens(self._convert_to_str(text))
        ret = []
        for token in words:
            ret.extend(self._token_to_subtoken_strings(token))
        return self._output_wrapper(ret, return_str=return_str)

    def detokenize(self, text, return_str=True):
        processed = False
        if isinstance(text, str):
            if " " in text:
                s = "".join(text.strip().split())
                processed = True
        if not processed:
            s = self._convert_to_str(text, delimiter="")
        ret = []
        for token in s.strip().split("_"):
            if token:
                unescaped_token = _unescape_token(token + "_")
                if unescaped_token:
                    ret.append(unescaped_token)
        ret = self.join_tokens_to_string(ret)
        return self._output_wrapper(ret, return_str=return_str)

    @staticmethod
    def init_from_files(vocab_file,
                        files,
                        target_vocab_size,
                        threshold,
                        min_count=None,
                        language="en",
                        file_byte_limit=1e6,
                        reserved_tokens=None):
        """Create subtoken vocabulary based on files, and save vocab to file.

        Args:
          vocab_file: String name of vocab file to store subtoken vocabulary.
          files: List of file paths that will be used to generate vocabulary.
          target_vocab_size: target vocabulary size to generate.
          threshold: int threshold of vocabulary size to accept.
          language: The language.
          min_count: int minimum count to use for generating the vocabulary. The min
            count is the minimum number of times a subtoken should appear in the
            files before it is added to the vocabulary. If set to none, this value
            is found using binary search.
          file_byte_limit: (Default 1e6) Maximum number of bytes of sample text that
            will be drawn from the files.
          reserved_tokens: List of string tokens that are guaranteed to be at the
            beginning of the subtoken vocabulary list.

        Returns:
          Subtokenizer object
        """
        if reserved_tokens is None:
            reserved_tokens = RESERVED_TOKENS

        if tf.io.gfile.exists(vocab_file):
            logging.info("Vocab file already exists (%s)", vocab_file)
        else:
            logging.info("Begin steps to create subtoken vocabulary...")
            token_counts = _count_tokens(files, file_byte_limit)
            alphabet = _generate_alphabet_dict(token_counts)
            subtoken_list = _generate_subtokens_with_target_vocab_size(
                token_counts, alphabet, target_vocab_size, threshold, min_count,
                reserved_tokens)
            logging.info("Generated vocabulary with %d subtokens.",
                         len(subtoken_list))
            _save_vocab_file(vocab_file, subtoken_list)
        subtokenizer = Subtokenizer(language)
        subtokenizer.init_subtokenizer(vocab_file)
        return subtokenizer



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from absl import app, logging

import neurst.utils.flags_core as flags_core
from neurst.utils.converters import Converter, build_converter

FLAG_LIST = [
    flags_core.Flag("from", dtype=flags_core.Flag.TYPE.STRING, default=None,
                    required=True, help="The path to pretrained model directory "
                                        "or a key indicating the publicly available model name."),
    flags_core.Flag("to", dtype=flags_core.Flag.TYPE.STRING, default=None,
                    required=True, help="The path to save the converted checkpoint."),
    flags_core.Flag("model_name", dtype=flags_core.Flag.TYPE.STRING, default=None,
                    required=True, help="The name of pretrained model, e.g. google_bert."),
]


def convert(converter: Converter, from_path, to_path):
    assert converter is not None
    assert from_path
    assert to_path
    converter.convert(from_path, to_path)


def _main(_):
    arg_parser = flags_core.define_flags(FLAG_LIST, with_config_file=False)
    args, remaining_argv = flags_core.parse_flags(FLAG_LIST, arg_parser)
    flags_core.verbose_flags(FLAG_LIST, args, remaining_argv)
    converter = build_converter(args["model_name"])
    convert(converter, args["from"], args["to"])


if __name__ == "__main__":
    logging.set_verbosity(logging.INFO)
    app.run(_main, argv=["pseudo.py"])



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from absl import logging

from neurst.utils.registry import REGISTRIES


def register_hparams_set(name):
    registry_name = "hparams_set"
    if registry_name not in REGISTRIES:
        REGISTRIES[registry_name] = {}

    def register_x_fn(fn_, short_name=None):
        names = set()
        if short_name:
            for n in short_name:
                names.add(n.lower())
        names.add(fn_.__name__)
        for n in names:
            if n in REGISTRIES[registry_name]:
                if REGISTRIES[registry_name][n] != fn_:
                    raise ValueError('Cannot register duplicate {} (under {})'.format(n, registry_name))
            else:
                REGISTRIES[registry_name][n] = fn_
    if isinstance(name, str):
        return lambda fn: register_x_fn(fn, [name])
    elif isinstance(name, list):
        return lambda c: register_x_fn(c, name)
    else:
        raise ValueError("Not supported type: {}".format(type(name)))


def get_hyper_parameters(name):
    registry_name = "hparams_set"
    if name is None:
        return {}
    if registry_name in REGISTRIES and name in REGISTRIES[registry_name]:
        logging.info("matched the pre-defined hyper-parameters set: {}".format(name))
        return REGISTRIES[registry_name][name]()
    for m, mc in REGISTRIES["model"].items():
        if hasattr(mc, "build_model_args_by_name"):
            p = mc.build_model_args_by_name(name)
            if p is not None:
                return p
    return {}



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta, abstractmethod

import six
import tensorflow as tf


@six.add_metaclass(ABCMeta)
class BaseModel(tf.keras.Model):
    REGISTRY_NAME = "model"

    def __init__(self, args, name=None):
        self._args = args
        super(BaseModel, self).__init__(name=name)

    @property
    def args(self):
        return self._args

    @staticmethod
    def class_or_method_args():
        return []

    @classmethod
    def new(cls, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def call(self, inputs, is_training=True):
        """ Forward pass of the model.

        Args:
            inputs: A dict of model inputs.
            is_training: A bool, whether in training mode or not.

        Returns:
            The model output.
        """
        raise NotImplementedError



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import time

import numpy
import tensorflow as tf
from absl import logging

from neurst.data.datasets.multiple_dataset import MultipleDataset
from neurst.data.datasets.text_gen_dataset import TextGenDataset
from neurst.exps import BaseExperiment, register_exp
from neurst.layers.search import SequenceSearch, build_search_layer
from neurst.metrics import Metric
from neurst.models.encoder_decoder_ensemble_model import EncoderDecoderEnsembleModel
from neurst.models.model_utils import summary_model_variables
from neurst.training import training_utils
from neurst.utils import compat
from neurst.utils.checkpoints import restore_checkpoint_if_possible
from neurst.utils.configurable import ModelConfigs
from neurst.utils.flags_core import Flag, ModuleFlag
from neurst.utils.misc import flatten_string_list, to_numpy_or_python_type


@register_exp(["predict", "generation"])
class SequenceGenerator(BaseExperiment):
    """ Entry for sequence generation. """

    def __init__(self, args, **kwargs):
        """ Initializes a util class for sequence generation. """
        super(SequenceGenerator, self).__init__(**kwargs)
        self._output_file = args["output_file"]
        self._save_metric = args["save_metric"]
        self._metric = self.task.get_eval_metric(args)
        self._search_layer = build_search_layer(args)

    @staticmethod
    def class_or_method_args():
        return [
            ModuleFlag(Metric.REGISTRY_NAME,
                       help="The evaluation metric for the generation results."),
            ModuleFlag(SequenceSearch.REGISTRY_NAME, help="The search layer for sequence generation."),
            Flag("output_file", dtype=Flag.TYPE.STRING, default=None,
                 help="The path to a file for generated outputs. If MultipleDataset is provided, "
                      "it should be a dict like {dataset_name0: data_path0, ...}"),
            Flag("save_metric", dtype=Flag.TYPE.STRING, default=None,
                 help="The path to a file that metrics will be saved to, in json format."),
        ]

    @staticmethod
    def build_generation_model(task, model, search_layer, output_sequence_only=True):
        """ Build keras model for generation.

        Args:
            task: The task object.
            model: An instance of neurst.models.model.BaseModel
            search_layer: A sequence search object.
            output_sequence_only: Only generated sequences will output if True.

        Returns: the generation model.
        """
        if search_layer is None:
            raise ValueError(
                "The parameters for generation method must be provided: "
                "search_method, search_method.params, ...")
        inps = task.create_inputs(compat.ModeKeys.INFER)
        formatted_inps = task.example_to_input(inps, compat.ModeKeys.INFER)
        search_layer.set_model(model)
        generation_ops = search_layer(formatted_inps)
        if output_sequence_only:
            generation_ops = generation_ops[0]
        keras_model = tf.keras.Model(inps, generation_ops)
        return keras_model

    def _build_and_restore_model(self):
        """ Build a single model or ensemble model. """
        model_dirs = flatten_string_list(self.model_dir)
        if len(model_dirs) == 1:
            model = self.model
            stat = restore_checkpoint_if_possible(model, model_dirs[0])
            if not stat:
                logging.info("WARNING: Fail to restore checkpoint from {}. "
                             "We assume this was done on purpose. ".format(model_dirs[0]))
        else:
            logging.info("We assume models for ensemble are all based on the same task.")
            multiple_models = []
            for idx, one_model_dir in enumerate(model_dirs):
                name_prefix = "ensemble_{}".format(idx)
                logging.info("Create model for {} from {}".format(name_prefix, one_model_dir))
                cfg = ModelConfigs.load(one_model_dir)
                this_model = self.task.build_model(cfg, name=name_prefix)
                stat = restore_checkpoint_if_possible(this_model, one_model_dir)
                if not stat:
                    logging.info("WARNING: Fail to restore checkpoint from {}. "
                                 "We assume this was done on purpose. ".format(one_model_dir))
                multiple_models.append(this_model)
            model = EncoderDecoderEnsembleModel.new(multiple_models)
        return model

    @staticmethod
    def postprocess_generation(task, generations):
        generations = numpy.concatenate(to_numpy_or_python_type(generations), 0)
        postprocess_fn = task.get_data_postprocess_fn(compat.ModeKeys.INFER)
        generations = [postprocess_fn(x) for x in generations]
        return generations

    def run(self):
        """ Sequence generation from an existing model checkpoint.

        Step 1: Build model and restore checkpoints.
        Step 2: Build test dataset.
        Step 3: Sequence generation.
        Step 4: Evaluation using metric.
        """
        # Step 3: Build model.
        with training_utils.get_strategy_scope(self.strategy):
            model = self._build_and_restore_model()
            keras_model = self.build_generation_model(self.task, model, self._search_layer)
            tfds = training_utils.build_datasets(compat.ModeKeys.INFER, self.strategy,
                                                 self.custom_dataset, self.task)
            keras_model.summary()
            summary_model_variables(keras_model)

        # Step 5: Sequence Generation.
        start_time = time.time()
        results = training_utils.make_predictions(
            self.strategy, keras_model, tfds, self.custom_dataset,
            map_func=lambda y: SequenceGenerator.postprocess_generation(self.task, y))
        logging.info("Generation elapsed: %.2fs", time.time() - start_time)

        if self._output_file:
            if isinstance(self.custom_dataset, MultipleDataset):
                if isinstance(self._output_file, dict):
                    for name in results:
                        if self._output_file.get(name, None):
                            with tf.io.gfile.GFile(self._output_file[name], "w") as fw:
                                fw.write("\n".join(results[name]) + "\n")
                            logging.info("Saving generation of dataset {} results into {}".format(
                                name, self._output_file[name]))
                else:
                    logging.info("Unsupported type of `output_file`={}({}) for MultipleDataset.".format(
                        self._output_file, type(self._output_file)))
            else:
                if isinstance(self._output_file, str):
                    with tf.io.gfile.GFile(self._output_file, "w") as fw:
                        fw.write("\n".join(results) + "\n")
                    logging.info("Saving generation results into {}".format(self._output_file))
                else:
                    logging.info(f"WARNING: No generation results are saved due to unsupported type "
                                 f"of `output_file`: {self._output_file} ({type(self._output_file)})")

        # Step 6: evaluation using metric
        def _display(res, name=None):
            if name:
                logging.info(f"Evaluation Result ({name}):")
            else:
                logging.info("Evaluation Result:")
            for k, v in res.items():
                logging.info("   %s=%.2f", k, v)

        if self._metric is not None:
            saving_metrics = dict()
            if isinstance(self.custom_dataset, MultipleDataset):
                on_average = {}
                mixed_dsnames = []
                mixed_hypos = []
                mixed_refs = []
                for name in tfds:
                    assert isinstance(self.custom_dataset.datasets[name], TextGenDataset)
                    if self.custom_dataset.datasets[name].targets:
                        metric_result = self._metric(results[name], self.custom_dataset.datasets[name].targets)
                        for k, v in metric_result.items():
                            if k not in on_average:
                                on_average[k] = 0.
                            on_average[k] += self.custom_dataset.sample_weights[name] * v
                        _display(metric_result, name)
                        mixed_dsnames.append(name)
                        mixed_hypos.extend(results[name])
                        mixed_refs.extend(self.custom_dataset.datasets[name].targets)
                        saving_metrics[name] = metric_result
                if len(mixed_dsnames) > 1:
                    _display(on_average, f"on average by weights {self._custom_dataset.sample_weights}")
                    mixed_metric_result = self._metric(mixed_refs, mixed_hypos)
                    _display(mixed_metric_result, "mixed of {}".format(",".join(mixed_dsnames)))
                    saving_metrics["MIXED"] = mixed_metric_result

            else:
                assert isinstance(self.custom_dataset, TextGenDataset)
                if self.custom_dataset.targets is not None:
                    metric_result = self._metric(results, self.custom_dataset.targets)
                    _display(metric_result)
                    saving_metrics = metric_result
            if self._save_metric is not None:
                logging.info(f"Saving metric results into {self._save_metric}")
                with tf.io.gfile.GFile(self._save_metric, "w") as fw:
                    json.dump(saving_metrics, fw)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math

import tensorflow as tf

from neurst.utils.activations import get_activation


class PrePostProcessingWrapper(tf.keras.layers.Layer):
    """ Custom prepost processing for transformer.

    The sequence is specified as a string which may contain the
    following characters:
      a: add previous_x
      n: apply normalization
      d: apply drop'out

    This class only defines the "n" - layer - "da" mode.
    """

    def __init__(self, layer, dropout_rate=0.1, epsilon=1e-12,
                 pre_norm=True, name="layer_prepostprocess"):
        """ Initializes.

        Args:
            layer: The layer.
            dropout_rate: The dropout rate.
            epsilon: The epsilon of layer norm.
            pre_norm: Applies norm->layer->dropout->residual if True, else
                layer->dropout->residual->norm
            name: The name of this layer.
        """
        super(PrePostProcessingWrapper, self).__init__(name=name)
        self._dropout_rate = dropout_rate
        self._epsilon = epsilon
        self._layer = layer
        self._norm_layer = None
        self._pre_norm = pre_norm

    def get_config(self):
        return dict(
            dropout_rate=self._dropout_rate,
            name=self.name)

    def build(self, input_shape):
        """ Creates norm layer. """
        self._norm_layer = tf.keras.layers.LayerNormalization(
            epsilon=self._epsilon, dtype="float32", name="ln")
        super(PrePostProcessingWrapper, self).build(input_shape)

    def call(self, inputs, *args, **kwargs):
        is_training = kwargs["is_training"]
        if self._pre_norm:
            # n
            y = self._norm_layer(inputs)
            # layer: self att / ffn
            y = self._layer(y, *args, **kwargs)
            # d
            if is_training:
                y = tf.nn.dropout(y, rate=self._dropout_rate)
            # a
            return inputs + y
        else:
            y = self._layer(inputs, *args, **kwargs)
            # d
            if is_training:
                y = tf.nn.dropout(y, rate=self._dropout_rate)
            # an
            return self._norm_layer(inputs + y)


class TransformerFFN(tf.keras.layers.Layer):
    """ Applies the position-wise feed-forward as described
    in https://arxiv.org/abs/1706.03762 """

    def __init__(self,
                 filter_size,
                 output_size,
                 dropout_rate,
                 activation="relu",
                 name="ffn"):
        """ Initializes Transformer FFN.

        Args:
            filter_size: The hidden size of the relu layer.
            output_size: The output size.
            dropout_rate: The dropout rate.
            activation: The activation of internal layer.
            name: The name of this layer.
        """
        super(TransformerFFN, self).__init__(name=name)
        self._dropout_rate = dropout_rate
        self._filter_size = filter_size
        self._output_size = output_size
        self._activation = activation
        self._activation_fn = get_activation(activation)
        self._conv1 = None
        self._conv2 = None

    def get_config(self):
        return dict(
            filter_size=self._filter_size,
            output_size=self._output_size,
            dropout_rate=self._dropout_rate,
            activation=self._activation,
            name=self.name)

    def build(self, input_shape):
        self._conv1 = tf.keras.layers.Dense(
            self._filter_size,
            activation=self._activation_fn,
            use_bias=True,
            name="dense1")
        self._conv2 = tf.keras.layers.Dense(
            self._output_size,
            activation=None,
            use_bias=True,
            name="dense2")
        super(TransformerFFN, self).build(input_shape)

    def call(self, inputs, is_training=False):
        """ Returns the output of TransformerFFN.

        Args:
            inputs: A tensor with shape [batch_size, length, num_units].
            is_training: A boolean scalar, whether in training mode or not.

        Returns:
            Output of the feedforward network.
            tensor with shape [batch_size, length, output_size]
        """
        output = self._conv1(inputs)
        if is_training:
            output = tf.nn.dropout(output, rate=self._dropout_rate)
        output = self._conv2(output)
        return output


class MultiHeadDenseLayer(tf.keras.layers.Layer):
    """ Auto splitting or combining heads for the linear transformation. """

    def __init__(self,
                 output_units,
                 num_heads,
                 kernel_initializer=None,
                 bias_initializer="zeros",
                 activation=None,
                 use_bias=True,
                 is_output_transform=False,
                 name="transform"):
        """ Initializes MultiHeadDenseLayer.

        Args:
            output_units: A int scalar or int list, indicating the transformed output units.
                It must be a int scalar when `is_output_transform` is True.
            num_heads: The head num.
            kernel_initializer: The initializer of kernel weight.
            bias_initializer: The initializer of bias.
            activation: A string or a callable function for activation.
            use_bias: A boolean, whether to add bias tensor.
            is_output_transform: A boolean, whether to use this layer for the output
                transformation in multi head attention.
            name: The name of the layer.
        """
        super(MultiHeadDenseLayer, self).__init__(name=name)
        self._output_units = output_units
        self._num_heads = num_heads
        self._kernel_initializer = kernel_initializer
        self._bias_initializer = bias_initializer
        self._use_bias = use_bias
        self._is_output_transform = is_output_transform
        self._activation = activation
        self._activation_fn = get_activation(activation)
        # compatible
        self._flatten_output_units = tf.nest.flatten(self._output_units)
        if is_output_transform:
            assert not tf.nest.is_nested(self._output_units)

    def get_config(self):
        return dict(
            output_units=self._output_units,
            num_heads=self._num_heads,
            kernel_initializer=self._kernel_initializer,
            bias_initializer=self._bias_initializer,
            activation=self._activation,
            use_bias=self._use_bias,
            is_output_transform=self._is_output_transform,
            name=self.name)

    def compat_kernel_shape(self, input_shape):
        """ Compatible kernel for variable storage. """
        if self._is_output_transform:
            return [input_shape[-1] * input_shape[-2], self._output_units]
        return [input_shape[-1], sum(self._flatten_output_units)]

    @property
    def kernel_shape(self):
        """ The kernel shape. """
        if self._is_output_transform:
            return [self._num_heads, -1, self._output_units]
        return [-1, sum(self._flatten_output_units)]

    @property
    def bias_shape(self):
        """ The bias shape. """
        return [sum(self._flatten_output_units)]

    def build(self, input_shape):
        """ Builds the layer. """
        self._kernel = self.add_weight(
            "kernel",
            shape=self.compat_kernel_shape(input_shape),
            initializer=self._kernel_initializer,
            trainable=True)
        self._bias = None
        if self._use_bias:
            self._bias = self.add_weight(
                "bias",
                shape=self.bias_shape,
                initializer=self._bias_initializer,
                trainable=True)
        super(MultiHeadDenseLayer, self).build(input_shape)

    def call(self, inputs):
        """ Implements ``call()`` for MultiHeadDenseLayer.

        Args:
            inputs: A float tensor of shape [batch_size, length, hidden_size]
                when output_projection is False, otherwise a float tensor of shape
                [batch_size, length, num_heads, num_units_per_head].

        Returns:
            The projected tensor with shape [batch_size, length, num_heads,
                num_units_per_head] per `self._output_units` when output_projection
                is False, otherwise [batch_size, length, output_units].
        """
        kernel = tf.keras.backend.reshape(self._kernel, self.kernel_shape)
        if self._is_output_transform:
            # a: batch
            # b: length
            # c: num heads
            # d: input units per head
            # e: num_output
            output = tf.einsum("abcd,cde->abe", inputs, kernel)
        else:
            # a: batch
            # b: length
            # c: input size
            # d: total output size
            output = tf.einsum("abc,cd->abd", inputs, kernel)
        if self._use_bias:
            output += self._bias

        if not self._is_output_transform:
            output = tf.split(
                output, self._flatten_output_units,
                axis=-1)
            output = tf.nest.map_structure(
                lambda x, num_units: tf.reshape(
                    x, tf.concat([tf.shape(x)[:-1],
                                  [self._num_heads, num_units // self._num_heads]], axis=0)),
                output, self._flatten_output_units)
        output = tf.nest.flatten(output)
        if self._activation_fn is not None:
            output = tf.nest.map_structure(
                self._activation_fn, output)
        return tf.nest.pack_sequence_as(self._output_units, output)


class PositionEmbeddingWrapper(tf.keras.layers.Layer):

    def __init__(self,
                 timing,
                 embedding_layer,
                 max_positions=512,
                 sinusoids_as_variable=False,
                 name="position_emb_wrapper"):
        """ Initializes the position embedding layer.

        Args:
            timing: The position embedding type. Now only 'sinusoids'
                and 'emb' are supported.
            embedding_layer: The embedding layer.
            max_positions: The maximum positions.
            sinusoids_as_variable: Whether the sinusoids position embedding
                is pre-calculated and fixed.
            name: The name of this layer.
        """
        super(PositionEmbeddingWrapper, self).__init__(name=name)
        self._timing = timing
        self._embedding_layer = embedding_layer
        self._embedding_dim = embedding_layer.embedding_dim
        self._max_positions = max_positions
        self._sinusoids_as_variable = sinusoids_as_variable
        assert self._timing in [None, "sinusoids", "emb"], (
            "Unknown position embedding type: \"{}\"".format(timing))

    @property
    def embedding_layer(self):
        return self._embedding_layer

    def get_config(self):
        return dict(
            timing=self._timing,
            max_positions=self._max_positions,
            embedding_dim=self._embedding_dim,
            name=self.name)

    def build(self, input_shape):
        with tf.name_scope("position_embeddings"):
            if self._timing == "emb":
                self._position_emb_table = self.add_weight(
                    "weights",
                    shape=[self._max_positions, self._embedding_dim],
                    initializer=tf.random_normal_initializer(
                        mean=0., stddev=self._embedding_dim ** -0.5),
                    trainable=True)
            elif self._timing == "sinusoids" and self._sinusoids_as_variable:
                self._position_emb_table = self.add_weight(
                    "weights",
                    shape=[self._max_positions, self._embedding_dim],
                    initializer=tf.constant_initializer(
                        self.add_sinusoids_timing_signal(
                            tf.zeros([1, self._max_positions, self._embedding_dim]), None).numpy()[0]),
                    trainable=False)
        super(PositionEmbeddingWrapper, self).build(input_shape)

    @staticmethod
    def add_sinusoids_timing_signal(x, time, min_timescale=1.0, max_timescale=1.0e4):
        """Adds a bunch of sinusoids of different frequencies to a Tensor.

        Each channel of the input Tensor is incremented by a sinusoid of a different
        frequency and phase.

        This allows attention to learn to use absolute and relative positions.
        Timing signals should be added to some precursors of both the query and the
        memory inputs to attention.

        The use of relative position is possible because sin(x+y) and cos(x+y) can be
        experessed in terms of y, sin(x) and cos(x).

        In particular, we use a geometric sequence of timescales starting with
        min_timescale and ending with max_timescale.  The number of different
        timescales is equal to channels / 2. For each timescale, we
        generate the two sinusoidal signals sin(timestep/timescale) and
        cos(timestep/timescale).  All of these sinusoids are concatenated in
        the channels dimension.

        This function is originally copied from Google's Tensor2Tensor code
        and modified to hold the capability for add timing signal at the
        specific time.

        Args:
          x: a Tensor with shape [batch, length, channels]
          min_timescale: a float
          max_timescale: a float

        Returns: A Tensor the same shape as x.
        """
        dtype = x.dtype.base_dtype
        channels = x.get_shape().as_list()[-1]
        if x.get_shape().ndims == 3:  # [batch_size, timesteps, dim]
            length = tf.shape(x)[1]
            position = tf.cast(tf.range(length), dtype=dtype)
        elif x.get_shape().ndims == 2:  # [batch_size, dim]
            length = 1
            position = tf.cast(tf.range(time, time + 1), dtype=dtype)
        else:
            raise ValueError("need a Tensor with rank 2 or 3")
        num_timescales = channels // 2
        log_timescale_increment = (
            math.log(float(max_timescale) / float(min_timescale))
            / (tf.cast(num_timescales, dtype) - 1))
        inv_timescales = min_timescale * tf.exp(
            tf.cast(tf.range(num_timescales), dtype=dtype) * -log_timescale_increment)
        scaled_time = tf.expand_dims(position, 1) * tf.expand_dims(inv_timescales, 0)
        signal = tf.concat([tf.sin(scaled_time), tf.cos(scaled_time)], axis=1)
        signal = tf.pad(signal, [[0, 0], [0, tf.math.mod(channels, 2)]])
        if x.get_shape().ndims == 3:
            signal = tf.reshape(signal, [1, length, channels])
        else:
            signal = tf.reshape(signal, [1, channels])
        return x + signal

    def call(self, inputs, time=None, **kwargs):
        emb = self._embedding_layer(inputs, **kwargs)
        mode = kwargs.get("mode", "embedding")
        if self._timing is None or mode != "embedding":
            return emb
        assert emb.get_shape()[-1] == self._embedding_dim, (
            "The position embedding dimension should match the "
            "embedding dimension: {} vs. {}".format(
                self._embedding_dim, emb.get_shape()[-1]))
        x_ndims = emb.get_shape().ndims
        if x_ndims == 2 and time is None:
            raise ValueError("\"time\" should be provided when input x has 2-dims")
        if self._timing == "sinusoids":
            emb *= self._embedding_dim ** 0.5
        # TO load from positional embedding from other repos, e.g. fairseq
        #     return self.add_sinusoids_timing_signal(
        #         x=emb, time=time)
        # if self._timing == "emb":
        if self._timing == "sinusoids" and not self._sinusoids_as_variable:
            return self.add_sinusoids_timing_signal(x=emb, time=time)
        if x_ndims == 2:
            position_emb = tf.gather(self._position_emb_table,
                                     tf.convert_to_tensor(time, dtype=tf.int32))
        elif x_ndims == 3:
            position_emb = tf.slice(self._position_emb_table, [0, 0],
                                    [tf.shape(emb)[1], -1])
        else:
            raise ValueError("need a Tensor with rank 2 or 3")
        return emb + tf.expand_dims(position_emb, axis=0)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from neurst.layers.common_layers import PositionEmbeddingWrapper
from neurst.layers.decoders.transformer_decoder import TransformerDecoder
from neurst.layers.modalities.text_modalities import WordEmbeddingSharedWeights
from neurst.models import register_model
from neurst.models.model import BaseModel
from neurst.utils.flags_core import Flag
from neurst.utils.hparams_sets import register_hparams_set


@register_model
class GPT2(BaseModel):
    """ Defines the GPT2 model. """

    def __init__(self,
                 args,
                 vocab_meta,
                 embedding,
                 decoder,
                 name=None):
        """ Initializes a GPT2 model.

        Args:
            args: A dict, containing the model configuration.
            vocab_meta: A dict containing vocabulary meta data, e.g. unk_id, mask_id, sep_id.
            embedding: The embedding layer (wrapped by positional encoding).
            decoder: The decoder.
            name: The name of the model.options = tf.data.Options()
        """
        super(GPT2, self).__init__(args, name=name or "gpt2")
        self._vocab_meta = vocab_meta
        self._embedding = embedding
        self._decoder = decoder
        self._output_linear_layer = None
        if not self._args["share_embedding_and_softmax_weights"]:
            self._output_linear_layer = tf.keras.layers.Dense(
                vocab_meta["vocab_size"], activation=None,
                use_bias=True, name="softmax_linear")

    @staticmethod
    def class_or_method_args():
        return [
            Flag("max_position_embeddings", dtype=Flag.TYPE.INTEGER, default=512,
                 help="The maximum numbers of positions."),
            Flag("num_layers", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of stacking layers of the decoder."),
            Flag("hidden_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of hidden units of the decoder."),
            Flag("num_attention_heads", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of heads of decoder self-attention."),
            Flag("filter_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of the filter size of decoder ffn."),
            Flag("ffn_activation", dtype=Flag.TYPE.STRING, default="gelu",
                 help="The activation function of decoder ffn layer."),
            Flag("attention_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of decoder self-attention layer."),
            Flag("attention_type", dtype=Flag.TYPE.STRING, default="dot_product",
                 help="The type of the attention function of decoder self-attention layer."),
            Flag("ffn_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of decoder ffn layer."),
            Flag("layer_postprocess_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate for each layer's post process in decoder."),
            Flag("layer_postprocess_epsilon", dtype=Flag.TYPE.FLOAT, default=1e-5,
                 help="The epsilon for layer normalization in decoder."),
        ]

    @classmethod
    def new(cls, args: dict, vocab_meta, name=None):
        """ Builds a sequence to sequence model.

        Args:
            args: A dict containing all model parameters.
            vocab_meta: A dict containing source-side vocabulary meta data, e.g. eos_id, vocab_size.
            name: The name of the model.

        Returns:
            A GPT2 model.
        """
        # build source and target modality
        embedding = PositionEmbeddingWrapper(
            embedding_layer=WordEmbeddingSharedWeights(
                embedding_dim=args["hidden_size"], vocab_size=vocab_meta["vocab_size"],
                share_softmax_weights=True, use_bias=False, name="embeddings"),
            name="posenc_wrapper", timing="emb", max_positions=args["max_position_embeddings"])
        decoder = TransformerDecoder(
            num_layers=args["num_layers"],
            hidden_size=args["hidden_size"],
            num_attention_heads=args["num_attention_heads"],
            filter_size=args["filter_size"],
            ffn_activation=args["ffn_activation"],
            attention_dropout_rate=args["attention_dropout_rate"],
            attention_type=args["attention_type"],
            ffn_dropout_rate=args["ffn_dropout_rate"],
            layer_postprocess_dropout_rate=args["layer_postprocess_dropout_rate"],
            layer_postprocess_epsilon=args["layer_postprocess_epsilon"],
            with_encoder_decoder_attention=False,
            name="decoder")
        model = cls(args, vocab_meta, embedding, decoder, name=name)
        _ = model({"tokens": tf.convert_to_tensor([[0, 1, 2]], tf.int64)})
        return model

    def output_logits_layer(self, features):
        """ Projects the decoder output to logits. """
        if self._output_linear_layer is None:
            return self._embedding(features, mode="linear")
        else:
            return self._output_linear_layer(features)

    def get_symbols_to_logits_fn(self, inputs, is_training, is_inference,
                                 decode_padded_length=None):
        """ Prepares for decoding.

        Args:
            inputs: A dict of model inputs.
                - tokens: int tensor with shape [batch_size, src_input_length].
            is_training: A bool, whether in training mode or not.
            is_inference: A bool, whether in generation mode or not.
            decode_padded_length: The maximum decoding length when inference, for creating
                static-shape cache.

        Returns:  A tuple of (decoding_internal_states, decoder_input, symbol_to_logit_fn)
        """
        tokens = inputs["tokens"]
        # [batch, length, hidden size]
        decoder_internal_cache = self._decoder.create_decoding_internal_cache(
            tokens, None, is_inference=is_inference,
            decode_padded_length=decode_padded_length)

        def symbols_to_logits_fn(symbols, cache, time=None):
            """ Generate logits for next potential IDs

            Args:
                symbols: Current decoded sequence.
                cache: A dictionary of values storing the previous decoder attention values.
                time: Loop index, or None for transformer training

            Returns: The logits Tensor.
            """
            inputs = self._embedding(symbols, time=time)
            if decode_padded_length is None:
                decoder_output = self._decoder(inputs, cache, is_training=is_training,
                                               decode_loop_step=None)
            else:
                decoder_output = self._decoder(inputs, cache, is_training=is_training,
                                               decode_loop_step=time)
            logits = self.output_logits_layer(decoder_output)
            return logits

        generation_initializer = {
            "decoder_input": inputs["tokens"],
            "decoder_internal_cache": decoder_internal_cache,
            "eos_id": self._vocab_meta["eos_id"],
            # "unk_id": self._vocab_meta["unk_id"]
        }
        return symbols_to_logits_fn, generation_initializer

    def call(self, inputs, is_training=True):
        """ Forward pass of the language model.

        Args:
            inputs: A dict of model inputs.
                - tokens: int tensor with shape [batch_size, length].
            is_training: A bool, whether in training mode or not.

        Returns:
            A logits Tensor.
        """
        symbols_to_logits_fn, generation_initializer = self.get_symbols_to_logits_fn(
            inputs, is_training=is_training, is_inference=False)
        return symbols_to_logits_fn(generation_initializer["decoder_input"],
                                    generation_initializer["decoder_internal_cache"])


def _gpt2_hparams(num_layers,
                  hidden_size,
                  filter_size,
                  dropout_rate,
                  epsilon,
                  num_heads,
                  max_positions):
    return {
        "model.class": GPT2.__name__,
        "model.params": {
            "share_embedding_and_softmax_weights": True,
            "max_position_embeddings": max_positions,
            "num_layers": num_layers,
            "hidden_size": hidden_size,
            "num_attention_heads": num_heads,
            "filter_size": filter_size,
            "ffn_activation": "gelu",
            "attention_dropout_rate": dropout_rate,
            "attention_type": "dot_product",
            "ffn_dropout_rate": dropout_rate,
            "layer_postprocess_dropout_rate": dropout_rate,
            "layer_postprocess_epsilon": epsilon
        }
    }


@register_hparams_set("gpt2_117m")
def gpt2_117m():
    return _gpt2_hparams(
        num_layers=12,
        hidden_size=768,
        filter_size=3072,
        dropout_rate=0.1,
        epsilon=1e-5,
        num_heads=12,
        max_positions=1024)


@register_hparams_set("gpt2_345m")
def gpt2_345m():
    return _gpt2_hparams(
        num_layers=24,
        hidden_size=1024,
        filter_size=4096,
        dropout_rate=0.1,
        epsilon=1e-5,
        num_heads=16,
        max_positions=1024)


@register_hparams_set("gpt2_toy")
def gpt2_toy():
    return _gpt2_hparams(
        num_layers=4,
        hidden_size=8,
        filter_size=16,
        dropout_rate=0.1,
        epsilon=1e-5,
        num_heads=2,
        max_positions=32)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from neurst.layers.encoders.transformer_encoder import TransformerEncoder
from neurst.layers.modalities.text_modalities import BertEmbedding
from neurst.models import register_model
from neurst.models.model import BaseModel
from neurst.utils.flags_core import Flag
from neurst.utils.hparams_sets import register_hparams_set


@register_model
class Bert(BaseModel):
    """ Defines the BERT model. """

    def __init__(self,
                 args,
                 vocab_meta,
                 embedding,
                 encoder,
                 pooler,
                 name=None):
        """ Initializes a BERT model.

        Args:
            args: A dict, containing the model configuration.
            vocab_meta: A dict containing vocabulary meta data, e.g. unk_id, mask_id, sep_id.
            word_embedding: The embedding layer (wrapped by positional encoding).
            token_type_embedding: The embedding layer of token types.
            encoder: The encoder.
            pooler: A dense layer that converters the encoded sequence tensor
                to a tensor of shape [batch, hidden].
            name: The name of the model.options = tf.data.Options()
        """
        super(Bert, self).__init__(args, name=name or "bert")
        self._vocab_meta = vocab_meta
        self._embedding = embedding
        self._encoder = encoder
        self._pooler = pooler

    @staticmethod
    def class_or_method_args():
        return [
            Flag("max_position_embeddings", dtype=Flag.TYPE.INTEGER, default=512,
                 help="The maximum numbers of positions."),
            Flag("num_layers", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of stacking layers of the encoder."),
            Flag("hidden_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of hidden units of the encoder."),
            Flag("num_attention_heads", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of heads of encoder self-attention."),
            Flag("filter_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of the filter size of encoder ffn."),
            Flag("ffn_activation", dtype=Flag.TYPE.STRING, default="gelu",
                 help="The activation function of encoder ffn layer."),
            Flag("attention_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of encoder self-attention layer."),
            Flag("attention_type", dtype=Flag.TYPE.STRING, default="dot_product",
                 help="The type of the attention function of encoder self-attention layer."),
            Flag("ffn_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of encoder ffn layer."),
            Flag("layer_postprocess_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate for each layer's post process in encoder."),
            Flag("layer_postprocess_epsilon", dtype=Flag.TYPE.FLOAT, default=1e-12,
                 help="The epsilon for layer normalization."),
        ]

    @classmethod
    def new(cls, args: dict, vocab_meta, name=None):
        """ Builds a sequence to sequence model.

        Args:
            args: A dict containing all model parameters.
            vocab_meta: A dict containing source-side vocabulary meta data, e.g. eos_id, vocab_size.
            name: The name of the model.

        Returns:
            A BERT model.
        """
        # build source and target modality
        embedding = BertEmbedding(
            embedding_dim=args["hidden_size"],
            vocab_size=vocab_meta["vocab_size"],
            max_positions=args["max_position_embeddings"],
            token_types=2,
            dropout_rate=args["ffn_dropout_rate"],
            epsilon=args["layer_postprocess_epsilon"],
            name="bert_embedding")
        encoder = TransformerEncoder(
            num_layers=args["num_layers"],
            hidden_size=args["hidden_size"],
            num_attention_heads=args["num_attention_heads"],
            filter_size=args["filter_size"],
            ffn_activation=args["ffn_activation"],
            attention_dropout_rate=args["attention_dropout_rate"],
            attention_type=args["attention_type"],
            ffn_dropout_rate=args["ffn_dropout_rate"],
            layer_postprocess_dropout_rate=args["layer_postprocess_dropout_rate"],
            layer_postprocess_epsilon=args["layer_postprocess_epsilon"],
            post_normalize=True, return_all_layers=True,
            name="encoder")
        pooler = tf.keras.layers.Dense(args["hidden_size"], activation=tf.nn.tanh,
                                       use_bias=True, name="pooler")
        model = cls(args, vocab_meta, embedding, encoder, pooler, name=name)
        _ = model({"tokens": tf.convert_to_tensor([[1, 2, 3]], tf.int64),
                   "padding": tf.convert_to_tensor([[0, 0., 0]], tf.float32)})
        return model

    def call(self, inputs, is_training=True):
        """ Forward pass of the BERT classifier.

        Args:
            inputs: A dict containing model inputs.
                - tokens: float tensor of shape [batch, length]
                - tokens_type: int tensor of shape [batch, length]
                - padding: int tensor of shape [batch, length], where 1 denotes the padding.
            is_training:

        Returns: A dict containing
            - encoder_output: [batch, length, hidden size]
            - pooled_output: [batch, hidden size]

        """
        tokens = inputs["tokens"]
        padding = inputs["padding"]
        tokens_type = inputs.get("tokens_type", None)
        embedded_inputs = self._embedding(tokens, tokens_type, is_training)
        # [batch, length, hidden size]
        encoder_outputs = self._encoder(embedded_inputs, padding, is_training=is_training)[-1]
        # We "pool" the model by simply taking the hidden state corresponding
        # to the first token. We assume that this has been pre-trained
        first_token_tensor = tf.squeeze(encoder_outputs[:, 0:1, :], axis=1)
        pooled_output = self._pooler(first_token_tensor)
        return {"encoder_outputs": encoder_outputs,
                "pooled_output": pooled_output}


def _bert_hparams(num_layers,
                  hidden_size,
                  filter_size,
                  dropout_rate,
                  num_heads,
                  max_positions):
    return {
        "model.class": Bert.__name__,
        "model.params": {
            "max_position_embeddings": max_positions,
            "num_layers": num_layers,
            "hidden_size": hidden_size,
            "num_attention_heads": num_heads,
            "filter_size": filter_size,
            "ffn_activation": "gelu",
            "attention_dropout_rate": dropout_rate,
            "attention_type": "dot_product",
            "ffn_dropout_rate": dropout_rate,
            "layer_postprocess_dropout_rate": dropout_rate,
            "layer_postprocess_epsilon": 1e-12,
        }
    }


@register_hparams_set("bert_base")
def bert_base():
    return _bert_hparams(
        num_layers=12,
        hidden_size=768,
        filter_size=3072,
        dropout_rate=0.1,
        num_heads=12,
        max_positions=512)


@register_hparams_set("bert_large")
def bert_large():
    return _bert_hparams(
        num_layers=24,
        hidden_size=1024,
        filter_size=4096,
        dropout_rate=0.1,
        num_heads=16,
        max_positions=512)


@register_hparams_set("bert_toy")
def bert_toy():
    return _bert_hparams(
        num_layers=4,
        hidden_size=8,
        filter_size=16,
        dropout_rate=0.1,
        num_heads=2,
        max_positions=32)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import numpy
import tensorflow as tf
from absl import logging

from neurst.models.wav2vec2 import Wav2Vec2
from neurst.utils.converters import Converter, register_converter
from neurst.utils.misc import download_with_tqdm

_URL_PREFIX = "https://dl.fbaipublicfiles.com/fairseq/wav2vec/"
_W2V_PRETRAIN_MODELS = {
    "wav2vec2_base": _URL_PREFIX + "wav2vec_small.pt",
    "wav2vec2_large": _URL_PREFIX + "libri960_big.pt",
}


@register_converter(["fairseq_wav2vec2", "wav2vec2"])
class FairseqWav2vec2(Converter):
    """ Converts from the fairseq's wav2vec2.0.
    https://github.com/pytorch/fairseq/tree/master/examples/wav2vec

    For fine-tuning usage, not pretraining.
    """

    @staticmethod
    def convert_model_config(path):
        import torch
        with tf.io.gfile.GFile(path, "rb") as fp:
            cp = torch.load(fp, map_location=torch.device('cpu'))
        args = cp["args"].__dict__
        activation_fn = args.get("activation_fn", "gelu")
        return {
            "model.class": Wav2Vec2.__name__,
            "model.params": {
                "conv_bias": args.get("conv_bias", False),
                "conv_feature_layers": args.get("conv_feature_layers",
                                                "[(512, 10, 5)] + [(512, 3, 2)] * 4 + [(512,2,2)] * 2"),
                "extractor_mode": args.get("extractor_mode", "default"),
                "encoder_embed_dim": args["encoder_embed_dim"],
                "conv_pos": args.get("conv_pos", 128),
                "conv_pos_groups": args.get("conv_pos_groups", 16),
                "dropout_input": args["dropout_input"],
                "encoder_layerdrop": args.get("encoder_layerdrop", 0.0),
                "encoder_dropout": args["dropout"],
                "encoder_layers": args["encoder_layers"],
                "encoder_attention_heads": args["encoder_attention_heads"],
                "encoder_filter_size": args["encoder_ffn_embed_dim"],
                "encoder_activation_fn": "gelu_nonapprox" if activation_fn == "gelu" else activation_fn,
                "encoder_pre_norm": args.get("layer_norm_first", False)
            }
        }

    @staticmethod
    def convert_task_config(path):
        raise NotImplementedError

    @staticmethod
    def download(key):
        if key in _W2V_PRETRAIN_MODELS:
            url = _W2V_PRETRAIN_MODELS[key]
        elif key.startswith("http://") or key.startswith("https://"):
            url = key
        else:
            return None
        model_file_name = os.path.basename(url)
        this_dir = os.path.dirname(__file__)
        extract_path = os.path.join(this_dir, model_file_name)
        if not os.path.exists(extract_path):
            logging.info(f"Downloading Wav2vec2.0: {key}")
            download_with_tqdm(url, extract_path)

        return extract_path

    @staticmethod
    def convert_checkpoint(path, save_path):
        import torch
        with tf.io.gfile.GFile(path, "rb") as fp:
            pyvars = torch.load(fp, map_location=torch.device('cpu'))["model"]
        cfgs: dict = FairseqWav2vec2.convert_model_config(path)["model.params"]
        py_model_prefix = ""
        for n in pyvars:
            if n.startswith("w2v_encoder.w2v_model."):
                py_model_prefix = "w2v_encoder.w2v_model."
                break
        # ========= feature extractor ========
        max_feconv_layers = 0
        fe_pyprefix = py_model_prefix + "feature_extractor.conv_layers"
        fe_tfprefix = "wav2vec2/feature_extractor"
        new_var_dict = dict()
        _name = f"{fe_pyprefix}.{max_feconv_layers}.0.weight"
        while _name in pyvars:
            _tfname = f"{fe_tfprefix}/conv_block{max_feconv_layers}/conv/kernel"
            new_var_dict[_tfname] = pyvars[_name].detach().numpy().transpose([2, 1, 0])
            max_feconv_layers += 1
            _name = f"{fe_pyprefix}.{max_feconv_layers}.0.weight"
        if cfgs["extractor_mode"] == "default":
            _name = f"{fe_pyprefix}.0.2."
            _tfname = f"{fe_tfprefix}/conv_block0/gn/"
            new_var_dict[_tfname + "gamma"] = pyvars[_name + "weight"].detach().numpy()
            new_var_dict[_tfname + "beta"] = pyvars[_name + "bias"].detach().numpy()
        else:
            raise NotImplementedError
        if cfgs["conv_bias"]:
            raise NotImplementedError
        _name = "wav2vec2/feature_extractor_norm/"
        new_var_dict[_name + "gamma"] = pyvars[py_model_prefix + "layer_norm.weight"].detach().numpy()
        new_var_dict[_name + "beta"] = pyvars[py_model_prefix + "layer_norm.bias"].detach().numpy()
        # post extract projection
        _name = "wav2vec2/post_extract_proj/"
        new_var_dict[_name + "kernel"] = pyvars[
            py_model_prefix + "post_extract_proj.weight"].detach().numpy().transpose()
        new_var_dict[_name + "bias"] = pyvars[py_model_prefix + "post_extract_proj.bias"].detach().numpy()
        # position conv
        _name = "wav2vec2/pos_conv/"
        new_var_dict[_name + "wn/g"] = pyvars[
            py_model_prefix + "encoder.pos_conv.0.weight_g"].detach().numpy()[0][0]
        new_var_dict[_name + "wn/kernel"] = pyvars[
            py_model_prefix + "encoder.pos_conv.0.weight_v"].detach().numpy().transpose([2, 1, 0])
        new_var_dict[_name + "wn/bias"] = pyvars[py_model_prefix + "encoder.pos_conv.0.bias"].detach().numpy()
        new_var_dict[_name + "ln/gamma"] = pyvars[py_model_prefix + "encoder.layer_norm.weight"].detach().numpy()
        new_var_dict[_name + "ln/beta"] = pyvars[py_model_prefix + "encoder.layer_norm.bias"].detach().numpy()
        # encoder
        for i in range(cfgs["encoder_layers"]):
            _name = f"wav2vec2/encoder/layer_{i}/"
            _pyname = py_model_prefix + f"encoder.layers.{i}."
            _selfatt_name = _name + "self_attention_prepost_wrapper/self_attention/"
            new_var_dict[_selfatt_name + "qkv_transform/kernel"] = numpy.concatenate([
                pyvars[_pyname + "self_attn.q_proj.weight"].detach().numpy().transpose(),
                pyvars[_pyname + "self_attn.k_proj.weight"].detach().numpy().transpose(),
                pyvars[_pyname + "self_attn.v_proj.weight"].detach().numpy().transpose()], axis=1)
            new_var_dict[_selfatt_name + "qkv_transform/bias"] = numpy.concatenate([
                pyvars[_pyname + "self_attn.q_proj.bias"].detach().numpy(),
                pyvars[_pyname + "self_attn.k_proj.bias"].detach().numpy(),
                pyvars[_pyname + "self_attn.v_proj.bias"].detach().numpy()], axis=0)
            new_var_dict[_selfatt_name + "output_transform/kernel"] = pyvars[
                _pyname + "self_attn.out_proj.weight"].detach().numpy().transpose()
            new_var_dict[_selfatt_name + "output_transform/bias"] = pyvars[
                _pyname + "self_attn.out_proj.bias"].detach().numpy()
            new_var_dict[_name + "self_attention_prepost_wrapper/ln/gamma"] = pyvars[
                _pyname + "self_attn_layer_norm.weight"].detach().numpy()
            new_var_dict[_name + "self_attention_prepost_wrapper/ln/beta"] = pyvars[
                _pyname + "self_attn_layer_norm.bias"].detach().numpy()
            _ffn_name = _name + "ffn_prepost_wrapper/ffn/"
            new_var_dict[_ffn_name + "dense1/kernel"] = pyvars[_pyname + "fc1.weight"].detach().numpy().transpose()
            new_var_dict[_ffn_name + "dense1/bias"] = pyvars[_pyname + "fc1.bias"].detach().numpy()
            new_var_dict[_ffn_name + "dense2/kernel"] = pyvars[_pyname + "fc2.weight"].detach().numpy().transpose()
            new_var_dict[_ffn_name + "dense2/bias"] = pyvars[_pyname + "fc2.bias"].detach().numpy()
            new_var_dict[_name + "ffn_prepost_wrapper/ln/gamma"] = pyvars[
                _pyname + "final_layer_norm.weight"].detach().numpy()
            new_var_dict[_name + "ffn_prepost_wrapper/ln/beta"] = pyvars[
                _pyname + "final_layer_norm.bias"].detach().numpy()

        ckpt_saver = tf.train.Checkpoint(**{
            name: tf.Variable(initial_value=numpy_var, trainable=False,
                              name=name, dtype="float32")
            for name, numpy_var in new_var_dict.items()})
        save_ckpt = os.path.join(save_path, "ckpt")
        logging.info(f"Saving checkpoint to {save_ckpt}")
        ckpt_saver.save(save_ckpt)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from neurst.data.data_pipelines import register_data_pipeline
from neurst.data.data_pipelines.data_pipeline import DataPipeline
from neurst.data.text.huggingface_tokenizer import HuggingFaceTokenizer
from neurst.data.text.vocab import PaddingMode, Vocab
from neurst.utils.converters.openai_gpt2 import OpenAIGPT2


@register_data_pipeline("gpt2")
class GPT2DataPipeline(DataPipeline, Vocab):

    def __init__(self,
                 language="en",
                 tokens=None,
                 vocab_path=None):
        """ Initializes the data pipeline from OpenAI released GPT-2.

        Args:
            language: The language.
            tokens: A list of word tokens.
            vocab_path: The path to the vocabulary file.
        """
        if tokens is None and vocab_path is None:
            path = OpenAIGPT2.download("117M")
            vocab_path = os.path.join(path, "encoder.json")
        Vocab.__init__(self, Vocab.load_tokens(vocab_path, tokens), lowercase=False)
        DataPipeline.__init__(self, language=language, tokens=self.tokens, vocab_path=None)
        self._language = language
        self._tokenizer = HuggingFaceTokenizer(language=language)
        self._tokenizer.init_subtokenizer("gpt2")
        self._eos_id = Vocab.map_token_to_id(self, "<|endoftext|>")

    @property
    def meta(self):
        return {
            "vocab_size": self.vocab_size,
            "eos_id": self._eos_id,
            "pad_id": self._eos_id,
            "padding_mode": PaddingMode.EOS_AS_PADDING,
            "language": self._language
        }

    def process(self, input, is_processed=False):
        """ Process one data sample.

        Args:
            input: A text string.
            is_processed: Whether the data sample is already processed.

        Returns:
            A list of generated token IDs.
        """
        if not is_processed:
            input = self._tokenizer.tokenize(input, return_str=False)
        elif isinstance(input, str):
            input = input.strip().split()
        token_ids = [x for x in Vocab.map_token_to_id(self, input) if x is not None]
        return token_ids + [self._eos_id]

    def recover(self, input):
        """ Recover one data sample.

        Args:
            input: A list of token ids, the output of neural model.

        Returns:
            A string, the recovered text.
        """
        try:
            eos_pos = input.index(self._eos_id)
            input = input[:eos_pos]
        except ValueError:
            pass
        output = Vocab.map_id_to_token(self, input)
        return self._tokenizer.detokenize(output, return_str=True)



import tensorflow as tf

from neurst.utils.registry import setup_registry

OPTIMIZER_REGISTRY_NAME = "optimizer"
build_optimizer, register_optimizer = setup_registry(OPTIMIZER_REGISTRY_NAME, base_class=tf.keras.optimizers.Optimizer,
                                                     verbose_creation=True)

Adam = tf.keras.optimizers.Adam
Adagrad = tf.keras.optimizers.Adagrad
Adadelta = tf.keras.optimizers.Adadelta
SGD = tf.keras.optimizers.SGD
register_optimizer(Adam)
register_optimizer(Adagrad)
register_optimizer(Adadelta)
register_optimizer(SGD)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta, abstractmethod

import six

from neurst.utils.configurable import extract_constructor_params


@six.add_metaclass(ABCMeta)
class DataPipeline(object):
    REGISTRY_NAME = "data_pipeline"

    def __init__(self, **kwargs):
        self._params = extract_constructor_params(locals(), verbose=False)

    def get_config(self) -> dict:
        return self._params

    @property
    @abstractmethod
    def meta(self) -> dict:
        """ The meta data. """
        return {}

    @abstractmethod
    def recover(self, input):
        """ Recovers one data sample. """
        raise NotImplementedError

    @abstractmethod
    def process(self, input, is_processed=False):
        """ Processes one data sample. """
        raise NotImplementedError



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np

from neurst.data.data_pipelines.transcript_data_pipeline import TranscriptDataPipeline
from neurst.metrics import register_metric
from neurst.metrics.metric import Metric


def _wer(ref, hypo):
    errors = np.zeros([len(ref) + 1, len(hypo) + 1, 3])
    errors[0, :, 1] = np.arange(len(hypo) + 1)
    errors[:, 0, 2] = np.arange(len(ref) + 1)
    substitution = np.array([1, 0, 0])
    insertion = np.array([0, 1, 0])
    deletion = np.array([0, 0, 1])
    for r, ref in enumerate(ref):
        for d, dec in enumerate(hypo):
            errors[r + 1, d + 1] = min((
                errors[r, d] + (ref != dec) * substitution,
                errors[r + 1, d] + insertion,
                errors[r, d + 1] + deletion), key=np.sum)

    return tuple(errors[-1, -1])


@register_metric
class Wer(Metric):
    def __init__(self, language="en", *args, **kwargs):
        _ = args
        _ = kwargs
        self._language = language
        super(Wer, self).__init__()

    def set_groundtruth(self, groundtruth):
        """ Setup inside groundtruth.

        Args:
            groundtruth: A list of references,
                [sent0_ref, sent1_ref, ...]
        """
        self._references = [TranscriptDataPipeline.cleanup_transcript(
            self._language, x, lowercase=True, remove_punctuation=True) for x in groundtruth]

    def greater_or_eq(self, result1, result2):
        return self.get_value(result1) <= self.get_value(result2)

    def get_value(self, result):
        if isinstance(result, (float, np.float32, np.float64)):
            return result
        return result["WER"]

    def call(self, hypothesis, groundtruth=None):
        """ Calculate wer

        Args:
            hypothesis: A list of hypothesis texts.
            groundtruth: A list of reference texts.

        Returns:
            A tuple(wer, substitutions, insertions, deletions)
        """
        if groundtruth is None:
            groundtruth = self._references
        else:
            groundtruth = [TranscriptDataPipeline.cleanup_transcript(
                self._language, x, lowercase=True, remove_punctuation=True) for x in groundtruth]
        hypothesis = [TranscriptDataPipeline.cleanup_transcript(
            self._language, x, lowercase=True, remove_punctuation=True) for x in hypothesis]
        substitutions = 0
        insertions = 0
        deletions = 0
        numwords = 0

        for lref, lout in zip(groundtruth, hypothesis):
            # read the reference and  output
            reftext, output = lref.strip().split(), lout.strip().split()

            # compare output to reference
            s, i, d = _wer(reftext, output)
            substitutions += s
            insertions += i
            deletions += d
            numwords += len(reftext)

        substitutions /= numwords
        deletions /= numwords
        insertions /= numwords
        error = substitutions + deletions + insertions
        return {
            "WER": error * 100.,
            "WER-substitutions": substitutions * 100.,
            "WER-insertions": insertions * 100.,
            "WER-deletions": deletions * 100.
        }



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta, abstractmethod

import six


@six.add_metaclass(ABCMeta)
class FeatureExtractor(object):
    """ Abstract feature extractor for extracting audio features. """
    REGISTRY_NAME = "feature_extractor"

    @abstractmethod
    def __call__(self, signal, rate):
        raise NotImplementedError

    @staticmethod
    def class_or_method_args():
        return []



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from neurst.layers.common_layers import MultiHeadDenseLayer
from neurst.utils.configurable import extract_constructor_params


class MultiHeadAttention(tf.keras.layers.Layer):
    """ Class of multi-head scaled-dot-product attention with input/output
        transformations. """

    def __init__(self,
                 num_heads,
                 num_units,
                 attention_key_depth=None,
                 attention_value_depth=None,
                 output_depth=None,
                 attention_dropout_rate=0.1,
                 attention_type="dot_product",
                 name=None):
        """ Initializes the multi head attention layer.

        Args:
            num_heads: A int scalar, the number of heads.
            num_units: A int scalar, the default units if other `depth` is
                not provided.
            attention_key_depth: A int scalar, the dimension for projected
                attention keys. If not provided, then use `num_units` as default.
            attention_value_depth: A int scalar, the dimension for projected
                attention values. If not provided, then use `num_units` as default.
            output_depth: A int scalar, the dimension for projected
                outputs. If not provided, then use `num_units` as default.
            attention_dropout_rate: A float scalar, the dropout rate for attention weight.
            attention_type: A string indicating the attention type.
            name: The name of the layer.
        """
        self._params = extract_constructor_params(locals(), verbose=False)
        super(MultiHeadAttention, self).__init__(name=name)
        self._num_heads = num_heads
        self._num_units = num_units
        self._attention_key_depth = attention_key_depth or num_units
        self._attention_value_depth = attention_value_depth or num_units
        self._output_depth = output_depth or num_units
        self._attention_dropout_rate = attention_dropout_rate
        self._attention_type = attention_type
        if self._attention_key_depth % self._num_heads != 0:
            raise ValueError("query depth ({}) must be divisible by the number of "
                             "attention heads ({}).".format(self._attention_key_depth, self._num_heads))
        if self._attention_value_depth % self._num_heads != 0:
            raise ValueError("value depth ({}) must be divisible by the number of "
                             "attention heads ({}).".format(self._attention_value_depth, self._num_heads))
        # pre-create output transform layer
        self._output_transform_layer = MultiHeadDenseLayer(
            output_units=self._output_depth, num_heads=self._num_heads,
            kernel_initializer="glorot_uniform", is_output_transform=True,
            use_bias=True, name="output_transform")

    def get_config(self):
        return self._params

    def build(self, input_shape):
        """ Builds the layer.
            Layers for linearly projecting the queries, keys, and values."""
        self._q_transform_layer = MultiHeadDenseLayer(
            output_units=self._attention_key_depth, num_heads=self._num_heads,
            kernel_initializer="glorot_uniform", is_output_transform=False,
            use_bias=True, name="q_transform")
        self._kv_transform_layer = MultiHeadDenseLayer(
            output_units=[self._attention_key_depth, self._attention_value_depth],
            num_heads=self._num_heads,
            kernel_initializer="glorot_uniform",
            is_output_transform=False,
            use_bias=True,
            name="kv_transform")
        self.built = True

    def compute_qkv(self, query, memory, cache, decode_loop_step=None):
        """ Computes linear transformations of query, keys and values.

        Args:
            query: A tensor with shape [batch_size, length_q, query_depth].
            memory: A tensor with shape [batch_size, length_m, memory_depth].
            cache: A dict, used during prediction.
            decode_loop_step: An integer, step number of the decoding loop. Used only
                for autoregressive inference with static-shape cache.

        Returns: A tuple `(query_transformed, key_transformed, memory_transformed)`.
        """
        _ = cache
        _ = decode_loop_step
        # [batch_size, length_q/k/v, num_heads, num_units_per_head]
        q = self._q_transform_layer(query)
        k, v = self._kv_transform_layer(memory)
        return q, k, v

    def att_fn(self, q, k, bias):
        """ Computes attention weights according to attention_type.

        Args:
            q: Attention query tensor with shape
              [batch_size, length_q, num_heads, att_key_depth / num_heads]
            k:  Attention query tensor with shape
              [batch_size, length_k, num_heads, att_key_depth / num_heads]
            bias: The bias tensor with shape [batch_size, length_q, length_k]
                or [batch_size, 1, length_q, length_k]

        Returns: The attention weight with shape
            [batch_size, num_heads, length_q, length_k]
        """
        if self._attention_type == "dot_product":
            # B: batch_size
            # T: length_k
            # F: length_q
            # N: num_heads
            # H: depth per head
            # logits: [batch_size, num_heads, length_q, length_k]
            logits = tf.einsum("BTNH,BFNH->BNFT", k, q)

            if bias is not None:
                if bias.get_shape().ndims == 2:
                    bias = tf.expand_dims(
                        tf.expand_dims(bias, axis=1), axis=1)
                elif bias.get_shape().ndims != 4:
                    raise ValueError("bias tensor with {}-dim is not valid".format(
                        bias.get_shape().ndims))
                logits += bias
            # Note that softmax internally performs math operations using float32
            # for numeric stability. When training with float16, we keep the input
            # and output in float16 for better performance.
            weights = tf.nn.softmax(logits)
        else:
            raise NotImplementedError(
                "att_fn for \"{}\" not implemented.".format(self._attention_type))
        return weights

    def call(self,
             query,
             memory,
             memory_bias=None,
             cache=None,
             is_training=True,
             decode_loop_step=None):
        """ Apply attention mechanism to query and memory.

        Args:
            query: A tensor with shape [batch_size, length_q, query_depth]
                or [batch_size, query_depth].
            memory: A tensor with shape [batch_size, length_m, memory_depth].
            memory_bias: A tensor with shape [batch_size, length_m],
                the attention bias that will be added to the result of the dot product.
            cache: (Used during prediction) A dictionary with tensors containing
                results of previous attentions. The dictionary must have the items:
                    {"keys": tensor with shape [batch_size, i, heads, dim_per_head],
                    "values": tensor with shape [batch_size, i, heads, dim_per_head]}
                where i is the current decoded length.
            is_training: A bool, whether in training mode or not.
            decode_loop_step: An integer, step number of the decoding loop. Used only
                for autoregressive inference with static-shape cache.

        Returns:
            Attention layer output with shape [batch_size, length_q, output_depth]
        """
        query_is_2d = False
        if query.get_shape().ndims == 2:
            # for using MultiHeadAttention in RNN-based decoders
            query_is_2d = True
            query = tf.expand_dims(query, axis=1)

        # linear transformation of q, k, v
        q, k, v = self.compute_qkv(query, memory, cache, decode_loop_step)

        # Scale query to prevent the dot product between query and key from growing
        q *= (self._attention_key_depth // self._num_heads) ** (-0.5)

        # compute attention weight, [batch_size, num_heads, length_q, length_k]
        weights = self.att_fn(q, k, memory_bias)
        if is_training:
            weights = tf.nn.dropout(weights, rate=self._attention_dropout_rate)
        # sum over attention values
        # N: num heads
        # F: length_q
        # T: length_k
        # H: num units per head
        # attention output: [batch_size, length_q, num_heads, num_units_per_head]
        attention_output = tf.einsum("BNFT,BTNH->BFNH", weights, v)

        # Run the outputs through another linear projection layer. Recombining heads
        # is automatically done --> [batch_size, length_q, num_units]
        attention_output = self._output_transform_layer(attention_output)
        if query_is_2d:
            # attention output: [batch_size, depth_value]
            attention_output = tf.squeeze(attention_output, axis=1)
        return attention_output


class MultiHeadSelfAttention(MultiHeadAttention):
    """ Class of multi-head scaled-dot-product self-attention with input/output
      transformations. """

    def build(self, input_shape):
        self._qkv_transform_layer = MultiHeadDenseLayer(
            output_units=[self._attention_key_depth,
                          self._attention_key_depth,
                          self._attention_value_depth],
            num_heads=self._num_heads,
            kernel_initializer="glorot_uniform",
            is_output_transform=False,
            use_bias=True,
            name="qkv_transform")
        self.built = True

    def call(self, query, bias=None, cache=None, is_training=True, decode_loop_step=None):
        """ Builds the self-attention context. """
        return super(MultiHeadSelfAttention, self).call(
            query=query,
            memory=query,
            memory_bias=bias,
            cache=cache,
            is_training=is_training,
            decode_loop_step=decode_loop_step)

    def compute_qkv(self, query, memory, cache, decode_loop_step=None):
        """ Computes linear transformations of query, keys and values, especially
            for self-attention in transformer.

        Args:
            query: Attention query tensor with shape [batch_size, length_q, channels_query],
                or [batch_size, 1, channels_query] for decoder self-attention
            memory: Unused.
            cache: Used during prediction.
            decode_loop_step: An integer, step number of the decoding loop. Used only
                for autoregressive inference with static-shape cache.

        Returns:
            A tuple `(query_transformed, key_transformed, memory_transformed)`.
        """
        _ = memory
        q, k, v = self._qkv_transform_layer(query)
        if cache is not None:
            # for self-attention in transformer decoder when mode=INFER
            if decode_loop_step is None:
                k = tf.concat([cache["keys"], k], axis=1)
                v = tf.concat([cache["values"], v], axis=1)
                cache["keys"] = k
                cache["values"] = v
            else:  # for dynamic shape

                def _insert_curr(_cache, _new_val):
                    size = _cache.get_shape().as_list()[1]
                    indices = tf.reshape(tf.one_hot(decode_loop_step, size, dtype=_new_val.dtype),
                                         [1, size, 1, 1])
                    new_val = _cache + _new_val * indices
                    return new_val

                cache["keys"] = _insert_curr(cache["keys"], k)
                cache["values"] = _insert_curr(cache["values"], v)
                k = cache["keys"][:, :decode_loop_step + 1]
                v = cache["values"][:, :decode_loop_step + 1]
        return q, k, v



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from absl import logging

from neurst.utils.configurable import deep_merge_dict

REGISTRIES = {}
REGISTRIED_CLS2ALIAS = {}


def setup_registry(registry_name, base_class=None, create_fn=None, verbose_creation=False):
    if registry_name not in REGISTRIES:
        REGISTRIES[registry_name] = {}
        REGISTRIED_CLS2ALIAS[registry_name] = {}

    from neurst.utils.flags_core import Flag, ModuleFlag

    def _verbose_creation(cls_, args, *extra_args, **kwargs):
        if not verbose_creation:
            return
        logging.info("Creating {}: {}".format(registry_name, cls_))
        if args is not None and len(args) > 0:
            logging.info("  ({}) arguments: ".format(registry_name))
            for k, v in args.items():
                if isinstance(v, dict):
                    logging.info("    {}:".format(k))
                    for kk, vv in v.items():
                        logging.info("      {}: {}".format(kk, vv))
                else:
                    logging.info("    {}: {}".format(k, v))
        if len(extra_args) > 0:
            logging.info("  ({}) extra args: ".format(registry_name))
            for x in extra_args:
                logging.info("    - {}".format(x))
        if len(kwargs) > 0:
            logging.info("  ({}) extra k-v args: ".format(registry_name))
            for k, v in kwargs.items():
                logging.info("    {}: {}".format(k, v))

    def build_x(args, *extra_args, **kwargs):
        params_ = {}
        if isinstance(args, dict):
            cls_ = (args.get("class", None)
                    or args.get(f"{registry_name}.class", None)
                    or args.get(f"{registry_name}", None))
            params_ = (args.get("params", None)
                       or args.get("{}.params".format(registry_name), {})) or {}
        else:
            cls_ = args
        if cls_ is None:
            return None
        if isinstance(cls_, str):
            if cls_.lower() == "none":
                return None
            if cls_ not in REGISTRIES[registry_name]:
                raise ValueError("Not registered class name: {}.".format(cls_))
            cls_ = REGISTRIES[registry_name][cls_]
            builder = cls_
        elif callable(cls_):
            builder = cls_
        else:
            raise ValueError("Not supported type: {} for builder.".format(type(cls_)))
        if create_fn is not None:
            assert hasattr(builder, create_fn), "{} has no {} for creation.".format(cls_, create_fn)
            builder = getattr(builder, create_fn)
        if kwargs is None:
            kwargs = {}
        assert isinstance(params_, dict), f"Not supported type: {type(params_)} for params"
        if hasattr(cls_, "class_or_method_args"):
            for f in cls_.class_or_method_args():
                if isinstance(f, Flag):
                    if f.name in kwargs:
                        params_[f.name] = kwargs.pop(f.name)
                    elif f.name not in params_:
                        params_[f.name] = f.default
                elif isinstance(f, ModuleFlag) and f.cls_key not in params_:
                    params_[f.cls_key] = f.default
                if isinstance(f, ModuleFlag) and f.params_key not in params_:
                    params_[f.params_key] = {}
            _verbose_creation(cls_, params_, *extra_args, **kwargs)
            return builder(params_, *extra_args, **kwargs)
        params_ = deep_merge_dict(params_, kwargs, merge_only_exist=False)
        _verbose_creation(cls_, {}, *extra_args, **params_)
        return builder(*extra_args, **params_)

    def register_x(name):

        def register_x_cls(cls_, short_name=None):
            if base_class is not None and not issubclass(cls_, base_class):
                raise ValueError('{} must extend {}'.format(cls_.__name__, base_class.__name__))
            names = set()
            if short_name:
                for n in short_name:
                    names.add(n)
            names.add(cls_.__name__)
            names.add(cls_.__name__.lower())
            names.add("_".join(re.sub("([A-Z])", r' \1', cls_.__name__).lower().strip().split()))
            for n in names:
                if n in REGISTRIES[registry_name]:
                    if REGISTRIES[registry_name][n] != cls_:
                        raise ValueError('Cannot register duplicate {} (under {})'.format(n, registry_name))
                else:
                    REGISTRIES[registry_name][n] = cls_
            REGISTRIED_CLS2ALIAS[registry_name][cls_.__name__] = names
            return cls_

        if isinstance(name, str):
            return lambda c: register_x_cls(c, [name])
        elif callable(name):
            return register_x_cls(cls_=name)
        elif isinstance(name, list):
            return lambda c: register_x_cls(c, name)
        else:
            raise ValueError("Not supported type: {}".format(type(name)))

    return build_x, register_x



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

import tensorflow as tf

from neurst.layers.decoders import build_decoder
from neurst.layers.encoders import build_encoder
from neurst.models import register_model
from neurst.models.encoder_decoder_model import EncoderDecoderModel
from neurst.utils.flags_core import Flag
from neurst.utils.hparams_sets import register_hparams_set


@register_model
class Transformer(EncoderDecoderModel):
    """ Defines the Transformer model. """

    def __init__(self, args, *largs, **kwargs):
        """ Initializes a sequence to sequence model.

        Args:
            args: A dict, containing the model configuration.
            src_meta: A dict containing source-side vocabulary meta data, e.g. eos_id, unk_id.
            trg_meta: A dict containing target-side vocabulary meta data, e.g. eos_id, unk_id.
            src_modality: The source side modality.
            trg_modality: The target side modality.
            encoder: The encoder.
            decoder: The decoder.
            name: The name of the model.options = tf.data.Options()
        """
        super(Transformer, self).__init__(args, *largs, **kwargs)
        self._args = args

    @staticmethod
    def class_or_method_args():
        this_args = [x for x in super(Transformer, Transformer).class_or_method_args() if x.name not in [
            "encoder.class", "decoder.class", "encoder.params", "decoder.params"]]
        this_args += [
            Flag("encoder.num_layers", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of stacking layers of the encoder."),
            Flag("encoder.hidden_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of hidden units of the encoder."),
            Flag("encoder.num_attention_heads", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of heads of encoder self-attention."),
            Flag("encoder.filter_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of the filter size of encoder ffn."),
            Flag("encoder.ffn_activation", dtype=Flag.TYPE.STRING, default="relu",
                 help="The activation function of encoder ffn layer."),
            Flag("encoder.attention_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of encoder self-attention layer."),
            Flag("encoder.attention_type", dtype=Flag.TYPE.STRING, default="dot_product",
                 help="The type of the attention function of encoder self-attention layer."),
            Flag("encoder.ffn_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of encoder ffn layer."),
            Flag("encoder.layer_postprocess_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate for each layer's post process in encoder."),
            Flag("encoder.post_normalize", dtype=Flag.TYPE.BOOLEAN, default=False,
                 help="Whether to apply layer norm after each encoder block."),
            Flag("encoder.layer_postprocess_epsilon", dtype=Flag.TYPE.FLOAT, default=1e-6,
                 help="The epsilon for layer normalization in encoder."),
            Flag("decoder.num_layers", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of stacking layers of the decoder."),
            Flag("decoder.hidden_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of hidden units of the decoder."),
            Flag("decoder.num_attention_heads", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of heads of decoder self-attention and encoder-decoder attention."),
            Flag("decoder.filter_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of the filter size of decoder ffn."),
            Flag("decoder.ffn_activation", dtype=Flag.TYPE.STRING, default="relu",
                 help="The activation function of decoder ffn layer."),
            Flag("decoder.attention_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of decoder self-attention and encoder-decoder attention."),
            Flag("decoder.attention_type", dtype=Flag.TYPE.STRING, default="dot_product",
                 help="The type of the attention function of decoder self-attention and encoder-decoder attention."),
            Flag("decoder.ffn_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of decoder ffn layer."),
            Flag("decoder.layer_postprocess_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate for each layer's post process in decoder."),
            Flag("decoder.post_normalize", dtype=Flag.TYPE.BOOLEAN, default=False,
                 help="Whether to apply layer norm after each decoder block."),
            Flag("decoder.layer_postprocess_epsilon", dtype=Flag.TYPE.FLOAT, default=1e-6,
                 help="The epsilon for layer normalization in decoder."),
        ]
        return this_args

    @classmethod
    def new(cls, args, src_meta, trg_meta, name=None):
        """ Builds a sequence to sequence model.

        Args:
            args: A dict containing all model parameters.
            src_meta: A dict containing source-side vocabulary meta data, e.g. eos_id, vocab_size.
            trg_meta: A dict containing target-side vocabulary meta data, e.g. eos_id, vocab_size.
            name: The name of the model.

        Returns:
            An encoder decoder model.
        """
        # build source and target modality
        src_modality, trg_modality = cls.build_modalities(args, src_meta, trg_meta)
        encoder_params, decoder_params = {}, {}
        for f in cls.class_or_method_args():
            if f.name in args:
                if f.name.startswith("encoder."):
                    encoder_params[f.name[8:]] = args[f.name]
                elif f.name.startswith("decoder."):
                    decoder_params[f.name[8:]] = args[f.name]
        # build encoder and decoder
        encoder = build_encoder({
            "encoder.class": "TransformerEncoder",
            "encoder.params": encoder_params})
        decoder = build_decoder({
            "decoder.class": "TransformerDecoder",
            "decoder.params": decoder_params})
        model = cls(args, src_meta, trg_meta, src_modality, trg_modality,
                    encoder, decoder, name=name)
        _ = model({"src": tf.convert_to_tensor([[1, 2, 3]], tf.int64),
                   "src_padding": tf.convert_to_tensor([[0, 0., 0]], tf.float32),
                   "trg_input": tf.convert_to_tensor([[1, 2, 3]], tf.int64)})
        return model

    @classmethod
    def build_model_args_by_name(cls, name):
        if name == "transformer_toy":
            dmodel = 8
            num_heads = 2
            num_encoder_layers = 2
            num_decoder_layers = 2
            num_encoder_filter_size = 10
            num_decoder_filter_size = 10
            dropout_rate = 0.1
        elif name == "transformer_base":
            dmodel = 512
            num_heads = 8
            num_encoder_layers = 6
            num_decoder_layers = 6
            num_encoder_filter_size = 2048
            num_decoder_filter_size = 2048
            dropout_rate = 0.1
        elif name == "transformer_s":
            dmodel = 256
            num_heads = 4
            num_encoder_layers = 6
            num_decoder_layers = 6
            num_encoder_filter_size = 2048
            num_decoder_filter_size = 2048
            dropout_rate = 0.1
        elif name == "transformer_big":
            dmodel = 1024
            num_heads = 16
            num_encoder_layers = 6
            num_decoder_layers = 6
            num_encoder_filter_size = 4096
            num_decoder_filter_size = 4096
            dropout_rate = 0.3
        elif name == "transformer_big_dp01":
            dmodel = 1024
            num_heads = 16
            num_encoder_layers = 6
            num_decoder_layers = 6
            num_encoder_filter_size = 4096
            num_decoder_filter_size = 4096
            dropout_rate = 0.1
        elif re.match(r"^transformer_\d+_\d+e_\d+d(_\d+h)?(_dp0\.\d+)?$", name):
            eles = name.split("_")
            dmodel = int(eles[1])
            num_encoder_layers = int(eles[2][:-1])
            num_decoder_layers = int(eles[3][:-1])
            num_heads = 8
            this_idx = 4
            if len(eles) > this_idx:
                if eles[this_idx][-1] == "h":
                    num_heads = int(eles[this_idx][:-1])
                    this_idx += 1
            assert dmodel % num_heads == 0, (
                "Invalid arguments in hparams_set: "
                "dimension({}) must be divisible by head({})".format(dmodel, num_heads))
            dropout_rate = 0.1
            if len(eles) > this_idx:
                if eles[this_idx][0:2] == "dp":
                    dropout_rate = float(eles[this_idx][2:])
            num_encoder_filter_size = dmodel * 4
            num_decoder_filter_size = dmodel * 4
        else:
            return None
        return {
            "model.class": cls.__name__,
            "model.params": {
                "modality.share_source_target_embedding": False,
                "modality.share_embedding_and_softmax_weights": True,
                "modality.dim": dmodel,
                "modality.timing": "sinusoids",
                "encoder.num_layers": num_encoder_layers,
                "encoder.hidden_size": dmodel,
                "encoder.num_attention_heads": num_heads,
                "encoder.filter_size": num_encoder_filter_size,
                "encoder.attention_dropout_rate": dropout_rate,
                "encoder.attention_type": "dot_product",
                "encoder.ffn_activation": "relu",
                "encoder.ffn_dropout_rate": dropout_rate,
                "encoder.post_normalize": False,
                "encoder.layer_postprocess_dropout_rate": dropout_rate,
                "decoder.num_layers": num_decoder_layers,
                "decoder.hidden_size": dmodel,
                "decoder.num_attention_heads": num_heads,
                "decoder.filter_size": num_decoder_filter_size,
                "decoder.attention_dropout_rate": dropout_rate,
                "decoder.attention_type": "dot_product",
                "decoder.ffn_activation": "relu",
                "decoder.ffn_dropout_rate": dropout_rate,
                "decoder.post_normalize": False,
                "decoder.layer_postprocess_dropout_rate": dropout_rate
            },
            "optimizer.class": "Adam",
            "optimizer.params": {
                "epsilon": 1.e-9,
                "beta_1": 0.9,
                "beta_2": 0.98
            },
            "lr_schedule.class": "noam",
            "lr_schedule.params": {
                "initial_factor": 1.0,
                "dmodel": dmodel,
                "warmup_steps": 4000
            },
        }


@register_hparams_set("transformer_toy")
def transformer_toy():
    return Transformer.build_model_args_by_name("transformer_toy")


@register_hparams_set("transformer_base")
def transformer_base():
    return Transformer.build_model_args_by_name("transformer_base")


@register_hparams_set("transformer_big")
def transformer_big():
    return Transformer.build_model_args_by_name("transformer_big")


@register_hparams_set("transformer_big_dp01")
def transformer_big_dp01():
    return Transformer.build_model_args_by_name("transformer_big_dp01")
